import numpy as np

from .base_model import BaseModel
from .loading.loading_factory import LoadingFactory
from .statistics.statistics_factory import StatisticsFactory
from ..settings.settings import Settings
from ...common.probability import ProbabilityFunctions


class WaterSystem(BaseModel):
    def __init__(self, settings: Settings):
        """
        Water System model
        """
        # Inherit
        super().__init__(settings)

        # Statistics
        self.statistics = StatisticsFactory.get_statistics(self.settings)

        # Loading
        self.loading = LoadingFactory.get_loading(self.settings)

    def process_slow_stochastics(
        self, exceedance_probability: np.ndarray, axis: int = None
    ) -> np.ndarray:
        """
        Convert probabilities given a discharge to probabilities per duration.

        This function converts probabilities represented by a discharge to
        probabilities per specified duration based on the given stochastic
        variable ('stochastic'). The probabilities are computed for each
        duration (basisduur).

        Parameters
        ----------
        exceedance_probability : np.ndarray
            Numpy array with the discharge on the first dimension.

        Returns
        -------
        np.ndarray
            Probabilities per specified duration (basisduur).
        """
        # Input stochastics
        statistics = self.get_statistics()
        slow_stochastics = list(statistics.stochastics_slow.keys())
        _axis = range(1, len(slow_stochastics) + 1) if axis is None else axis

        # Put slow variable on first axis if not already
        newpos = range(len(np.atleast_1d(_axis)))
        if newpos != _axis:
            ep_x = np.moveaxis(exceedance_probability, _axis, newpos)
        else:
            ep_x = exceedance_probability

        # Calculate exceedance frequencies per peak discharge / lake level
        ep_ks = self.process_wave_shape(ep_x)

        # Steps in probability density
        if len(slow_stochastics) == 1 and slow_stochastics[0] == "q":
            p_peak = ProbabilityFunctions.probability_density(
                statistics.discharge.qpeak, statistics.discharge.epqpeak
            ).probability

        elif len(slow_stochastics) == 1 and slow_stochastics[0] == "a":
            p_peak = ProbabilityFunctions.probability_density(
                statistics.lake_level.apeak, statistics.lake_level.epapeak
            ).probability

        elif len(slow_stochastics) == 2 and slow_stochastics == ["a", "q"]:
            dq = ProbabilityFunctions.probability_density(
                statistics.discharge.qblok, statistics.discharge.epqpeak
            ).delta
            dm = ProbabilityFunctions.probability_density(
                statistics.lake_level.ablok, statistics.lake_level.epapeak
            ).delta
            p_peak = statistics.density_aq_peak * dm[:, None] * dq[None, :]

        elif len(slow_stochastics) == 2 and slow_stochastics == ["q", "a"]:
            dq = ProbabilityFunctions.probability_density(
                statistics.discharge.qblok, statistics.discharge.epqpeak
            ).delta
            da = ProbabilityFunctions.probability_density(
                statistics.lake_level.ablok, statistics.lake_level.epapeak
            ).delta
            p_peak = statistics.density_aq_peak.T * dq[:, None] * da[None, :]

        # Reshape steps when multiplying to trapezoidal probs
        shp = p_peak.shape + (1,) * (ep_ks.ndim - p_peak.ndim)

        # Trapezoidal probability
        p_trapezoidal = (p_peak.reshape(shp) * ep_ks).sum(
            axis=tuple(range(p_peak.ndim))
        )

        return p_trapezoidal

    def process_wave_shape(self, exceedance_probability: np.ndarray) -> np.ndarray:
        """
        Process the discharge waveshapes into the exceedance probabilities
        conditioned on the discharge level. The discharge level is represented
        on the first axis, which has dimensions equal to the number of
        discharge steps in the statistics.

        Parameters
        ----------
        exceedance_probability : np.ndarray
            Numpy array containing the exceedance probabilities with the
            discharge on the first dimension.

        Returns
        -------
        blokondkans : np.ndarray
            Exceedance probabilities of the discharge per duration (basisduur).
        """
        # Input variables
        statistics = self.get_statistics()
        slow_stochastics = list(statistics.stochastics_slow.keys())
        fp = exceedance_probability

        # Result variables
        intidxs = []
        xs = []
        xps = []

        # Verzamel de interpolatieindices voor alle trage stochasten
        for i, _stochastic in enumerate(slow_stochastics):
            if _stochastic == "q":
                xp = statistics.discharge.qpeak
                x = statistics.discharge.get_wave_shape().get_wave_shapes()
                tijden = statistics.discharge.get_wave_shape().time
            elif _stochastic == "a":
                xp = statistics.lake_level.apeak
                x = statistics.lake_level.get_wave_shape().get_wave_shapes()
                tijden = statistics.lake_level.get_wave_shape().time
            else:
                raise ValueError(
                    f"Eerste trage stochast is niet 'q' of 'm'/'a' (gegeven eerste trage stochast: {_stochastic})"
                )

            # Check if the length matches
            if exceedance_probability.shape[i] != x.shape[1]:
                raise ValueError(
                    f"The number of elements in the stochast axis ({i}) should be {x.shape[1]} ({x.shape}), but is {exceedance_probability.shape[i]} ({exceedance_probability.shape})."
                )

            # Bepaal de interpolatie-indices, de xp-punten rondom x waartussen wordt geïnterpoleerd
            intidx = np.array(
                [(xp[None, :] <= ix[:, None]).sum(1) - 1 for ix in x], dtype=np.uint16
            )
            intidx = np.minimum(np.maximum(intidx, 0), len(xp) - 2)

            intidxs.append(intidx)
            xs.append(x)
            xps.append(xp)

        # Bepaal de tijdsduur van het eerste en het laatste blokje uit de golfvorm
        block_duration = tijden[1] - tijden[0]
        eindduur = tijden[-1] - tijden[-2]

        # Alloceer arrays. Het is sneller om in de loop de arrays over te schrijven dan om ze opnieuw aan te maken
        ondkanstot = np.ones(exceedance_probability.shape)
        blokondkans = np.zeros(exceedance_probability.shape)
        ovkansen = np.zeros(exceedance_probability.shape)

        nt = len(tijden)

        for it in range(nt - 1):
            # Interpoleer de overschrijdingskansen
            if len(slow_stochastics) == 1:
                # Bepaal de index van de interpolatiewaarden (afvoeren / meerpeil) voor de betreffende tijdstap
                iix = intidxs[0][it]

                # Bepaal de fracties van de fp waarden obv xp, die moet worden ingevuld voor f (obv x)
                fracs = (x[it] - xp[iix]) / (xp[iix + 1] - xp[iix])

                # Bepaal de vorm van de output array, dit is (Ntraag, Ntijd, Nderest1, Nderest2, ...)
                fracshp = fracs.shape + (1,) * (fp.ndim - 1)

                # Interpoleer alle overschrijdingskansen in één keer door de fracties met de resultaatwaarden (fp) te vermenigvuldigen
                ovkansen[:] = (1 - fracs.reshape(fracshp)) * fp[iix] + fp[
                    iix + 1
                ] * fracs.reshape(fracshp)

            elif len(slow_stochastics) == 2:
                iix1 = intidxs[0][it]
                iix2 = intidxs[1][it]

                # Bepaal de fracties van de fp waarden obv xp, die moet worden ingevuld voor f (obv x)
                fracs1 = (xs[0][it] - xps[0][iix1]) / (xps[0][iix1 + 1] - xps[0][iix1])
                fracs2 = (xs[1][it] - xps[1][iix2]) / (xps[1][iix2 + 1] - xps[1][iix2])

                # Bepaal de vorm van de output array, dit is (Ntraag, Ntijd, Nderest1, Nderest2, ...)
                frac2shp = fracs2.shape + (1,) * (fp.ndim - 2)

                # 2D interpolatie
                # Merk op dat dit via een loop gaat. Dit is sneller dan in enkele array-operaties, omdat
                # de arrays zo groot zijn dat fancy-indexing traag wordt
                for i, (ix, ixp1) in enumerate(zip(iix1, iix1 + 1)):
                    # Interpoleer eerst over de eerste stochast
                    fy1 = (1 - fracs1[i]) * fp[ix][iix2] + fp[ixp1][iix2] * fracs1[i]
                    fy2 = (1 - fracs1[i]) * fp[ix][iix2 + 1] + fp[ixp1][
                        iix2 + 1
                    ] * fracs1[i]

                    # Interpoleer de tweede stochast
                    ovkansen[i] = fy1 + (fy2 - fy1) * fracs2.reshape(frac2shp)

                # fy1 = (1 - fracs1.reshape(frac1shp)) * fp[np.ix_(iix1, iix2  )] + fp[np.ix_(iix1+1, iix2  )] * fracs1.reshape(frac1shp)
                # fy2 = (1 - fracs1.reshape(frac1shp)) * fp[np.ix_(iix1, iix2+1)] + fp[np.ix_(iix1+1, iix2+1)] * fracs1.reshape(frac1shp)

            # Door het interpoleren kunnen overschrijdingskansen gecreëerd worden, die een fractie groter zijn dan 1.0
            ovkansen[:] = np.minimum(ovkansen, 1.0)

            # De berekende overschrijdingskansen gelden voor een blokje met de duur, die gelijk is
            # aan de door de gebruiker gekozen block_duration van de wind. Voor het discretiseren van de
            # golfvormen heeft de gebuiker een andere duur opgegeven. De bovenstaande overschrijdings-
            # kansen worden hiervoor gecorrigeerd
            if block_duration != statistics.wind_speed.block_duration_wind:
                ovkansen[:] = 1.0 - (1.0 - ovkansen) ** (
                    block_duration / statistics.wind_speed.block_duration_wind
                )

            # Bereken de onderschrijdingskans voor de basisduur gebruikmakend van de overschrijdingskansen per blokje
            # Factoren zijn 1
            blokondkans[:] = 1.0 - ovkansen

            # Eerste blok is half
            if it == 0:
                blokondkans[:] **= 0.5

            # Neem het voorlaatste en laatste blok samen
            if it == (nt - 1):
                # Als de eindduur kleiner is dan een halve block_duration, dan is de factor kleiner dan 1
                if eindduur < (0.5 * block_duration):
                    factor = ((0.5 * block_duration) + eindduur) / block_duration
                    blokondkans[:] **= factor

                # Als de eindduur groter of even lang is als een halve block_duration
                else:
                    # Vermenig de block_duration ook nog met het laatste blokje
                    factor = (eindduur - (block_duration / 2.0)) / block_duration
                    blokondkans[:] *= (1.0 - ovkansen[:, :]) ** factor

            ondkanstot *= blokondkans

        # Return
        return 1 - ondkanstot



from abc import ABC, abstractmethod

from .stochastics.barrier.barrier import Barrier
from .stochastics.barrier.no_barrier import NoBarrier
from .stochastics.discharge import Discharge
from .stochastics.discrete_probability import DiscreteProbability
from .stochastics.lake_level import LakeLevel
from .stochastics.model_uncertainty import ModelUncertainty
from .stochastics.sea_level.sea_level import SeaLevel
from .stochastics.sigma_function import SigmaFunction

# from .stochastics.wave_shape import WaveShape
from .stochastics.wind_speed import WindSpeed
from ...settings.settings import Settings


class Statistics(ABC):
    """
    Statistics abstract base class
    Used to define statistic classes for different water systems
    """

    @abstractmethod
    def __init__(self, settings: Settings):
        # Save settings
        self.settings: Settings = settings

        # Statistics, required
        self.wind_direction: DiscreteProbability = None
        self.wind_speed: WindSpeed = None
        self.model_uncertainties: ModelUncertainty = None

        # Coast
        self.barrier: Barrier = NoBarrier()
        self.phase_surge_tide: DiscreteProbability = None
        self.sea_level: SeaLevel = None
        self.storm_surge_duration: DiscreteProbability = None

        # Lake
        self.lake_level: LakeLevel = None

        # Other
        self.sigma_function: SigmaFunction = None

        # River
        self.discharge: Discharge = None

        # Discrete, slow, fast stochatics
        # TODO: Replace with framework
        self.stochastics_discrete = {}
        self.stochastics_fast = {}
        self.stochastics_slow = {}

    @abstractmethod
    def calculate_probability(
        self, wind_direction: float, closing_situation: int = 1, given: list = []
    ):
        """
        TODO
        """
        pass

    def get_wind_directions(self) -> DiscreteProbability:
        """
        Returns the wind directions statistics object.

        Returns
        -------
        WindDirection
            The wind direction statistics
        """
        return self.wind_direction

    def get_wind_speed(self) -> WindSpeed:
        """
        Returns the wind statistics object.

        Returns
        -------
        WindSpeed
            The wind speed statistics
        """
        return self.wind_speed

    def get_model_uncertainties(self) -> ModelUncertainty:
        """
        Returns the model uncertainty statistics object.

        Returns
        -------
        ModelUncertainty
            The ModelUncertainty statistics
        """
        return self.model_uncertainties

    def get_barrier(self) -> Barrier:
        """
        Returns the barrier statistics object.

        Returns
        -------
        Barrier
            The barrier statistics object
        """
        return self.barrier

    def get_discharge(self) -> Discharge:
        """
        Returns the discharge statistics object.

        Returns
        -------
        Discharge
            The discharge statistics object
        """
        return self.discharge

    def get_lake_level(self) -> LakeLevel:
        """
        Returns the lake level statistics object.

        Returns
        -------
        LakeLevel
            The LakeLevel statistics
        """
        return self.lake_level

    def get_phase_surge_tide(self) -> DiscreteProbability:
        """
        Returns the fase surge tide statistics object.

        Returns
        -------
        FaseSurgeTide
            The FaseSurgeTide statistics
        """
        return self.phase_surge_tide

    def get_sea_level(self) -> SeaLevel:
        """
        Returns the sea level statistics object.

        Returns
        -------
        SeaLevel
            The SeaLevel statistics
        """
        return self.sea_level

    def get_sigma_function(self) -> SigmaFunction:
        """
        Returns the sigma function statistics object.

        Returns
        -------
        SigmaFunction
            The SigmaFunction statistics
        """
        return self.sigma_function

    def get_storm_surge_duration(self) -> DiscreteProbability:
        """
        Returns the storm surge duration statistics object.

        Returns
        -------
        StormSurgeDuration
            The StormSurgeDuration statistics
        """
        return self.storm_surge_duration



import numpy as np

from collections import namedtuple

pdstruct = namedtuple("pdstruct", ["delta", "probability", "density", "edges"])


class ProbabilityFunctions:
    """
    A class with common functions used within statistics.
    """

    @staticmethod
    def probability_density(
        values: np.ndarray,
        exceedance_probability: np.ndarray,
        bounded: bool = True,
        check: bool = True,
        axis: int = None,
    ) -> pdstruct:
        """
        Function to convert the exceedance probability into a probability
        density function.

        Parameters
        ----------
        values : np.ndarray
            Values, for example the wind speed discretisation
        exceedance_probability : np.ndarray
            Exceedance probability of the values
        bounded : bool, optional
            If bounded, add the first and last element based on the min and max
            in the exceedance probability. Otherwise between 1.0 and 0.0
            (default : True)
        check : bool, optional
            Check whether the values and exceedance probabilities are
            monotonously increasing (default : True)
        axis : int, optional
            Axis (default : None)

        Raises
        ------
        ValueError
            If the values or exceedance probabilities are not monotonously
            increasing

        Returns
        -------
        pdstruct
            Probability density structure
        """
        # For multiple dimensions, use the _nd function
        if exceedance_probability.ndim > 1:
            return ProbabilityFunctions.probability_density_nd(
                values, exceedance_probability, bounded, axis=axis
            )

        # Check whether the values and exceedance probabilities are monotonously increasing
        if check:
            for arr, tag in zip(
                [values, exceedance_probability], ["Values", "Exceedance probabilities"]
            ):
                diff = arr[1:] - arr[:-1]
                if not (all(diff >= 0) or all(diff <= 0)):
                    raise ValueError(
                        tag + "are not monotonously increasing or decreasing.",
                        arr,
                        diff,
                    )

        # Determine the exceedance probability bins
        bins_edges = (exceedance_probability[1:] + exceedance_probability[:-1]) / 2.0

        # If bounded, add the first and last element based on the min and max in the exceedance probability
        if bounded:
            bins_edges = np.concatenate(
                [[exceedance_probability[0]], bins_edges, [exceedance_probability[-1]]]
            )

        # Else, determine the bins between the 0 and 1
        else:
            if exceedance_probability[0] < exceedance_probability[-1]:
                bins_edges = np.concatenate([[0.0], bins_edges, [1.0]])
            else:
                bins_edges = np.concatenate([[1.0], bins_edges, [0.0]])

        # The difference between the bin_edges are the bins_probabilities
        bins_probability = np.absolute(bins_edges[1:] - bins_edges[:-1])

        # Determine the delta of the values
        bins_values = np.concatenate(
            [[values[0]], (values[1:] + values[:-1]) / 2.0, [values[-1]]]
        )
        bins_deltas = np.absolute(np.diff(bins_values))

        # Probability density is the bins_probability divided by the delta
        probability_density = np.absolute(bins_probability / bins_deltas)

        # Return as a structure
        return pdstruct(bins_deltas, bins_probability, probability_density, bins_edges)

    @staticmethod
    def probability_density_nd(
        values: np.ndarray,
        exceedance_probability: np.ndarray,
        bounded: bool = True,
        axis: int = None,
    ) -> pdstruct:
        """
        Convert the exceedance probability into a probability density function
        for multidimensional arrays.

        This function converts the exceedance probabilities
        (exceedance_probability) into a probability density function (PDF) for
        multidimensional arrays represented by 'values' and
        'exceedance_probability'.

        Parameters
        ----------
        values : np.ndarray
            Values, e.g., wind speed discretization, for which the PDF is
            calculated.
        exceedance_probability : np.ndarray
            Exceedance probability of the values.
        bounded (bool):
            If True, add the first and last element based on the min and max in
            the exceedance probability. Otherwise, create the PDF between 1.0
            and 0.0.
        axis : int, optional
            The axis along which the PDF is calculated. By default, axis 0 is
            used.

        Returns
        -------
        pdstruct
            Probability density structure containing the calculated PDF.
        """
        # If bounded, add the first and last element based on the min and max
        # in the exceedance probability
        if bounded:
            bins_edges = np.concatenate(
                [
                    [exceedance_probability[0]],
                    (exceedance_probability[1:] + exceedance_probability[:-1]) / 2.0,
                    [exceedance_probability[-1]],
                ]
            )

        # If not
        else:
            bins_edges = np.pad(
                (exceedance_probability[1:] + exceedance_probability[:-1]) / 2,
                pad_width=(1, 1),
                mode="constant",
                constant_values=(0, 1)
                if (exceedance_probability[0] < exceedance_probability[-1]).all()
                else (1, 0),
            )

        # The difference between the bin_edges are the bins_probabilities
        bins_probability = np.absolute(bins_edges[1:, ...] - bins_edges[:-1, ...])

        # Edges between consecutive values, the difference gives the bin_deltas
        bins_deltas = np.absolute(
            np.diff(
                np.concatenate(
                    [[values[0]], (values[1:] + values[:-1]) / 2.0, [values[-1]]]
                )
            )
        )

        # Probability density is the bins_probability divided by the delta
        shp = [1] * bins_probability.ndim
        if axis is None:
            axis = 0
        shp[axis] = -1
        probability_density = np.absolute(
            bins_probability / bins_deltas.reshape(tuple(shp))
        )

        # Return as a structure
        return pdstruct(bins_deltas, bins_probability, probability_density, bins_edges)

    @staticmethod
    def get_hnl_disc_array(vmin: float, vmax: float, step: float) -> np.ndarray:
        """
        Get a discretized array of values between vmin and vmax with the given
        step size.

        This function generates a discretized array of values in the specified
        range [vmin, vmax] with a given step size. The array includes the vmin
        and vmax values and is uniformly spaced with steps of the specified
        size.

        Parameters
        ----------
        vmin : float
            Minimum value of the range.
        vmax : float
            Maximum value of the range.
        step : float
            Step size between values.

        Returns
        -------
        np.array
            Discretized array of values.
        """
        # Calculate the number of steps between vmin and vmax and create the
        # discretized array
        n = round((vmax - vmin) / step)
        levels = np.arange(vmin, vmin + (n + 0.1) * step, step)

        # Ensure that the last element of the array is exactly equal to vmax
        levels[-1] = vmax
        return levels

    @staticmethod
    def conditional_probability(probability: np.ndarray, axis: int) -> np.ndarray:
        """
        Calculate the conditional probability along an axis, taking into account
        dividing by zero.

        This function calculates the conditional probability along the
        specified 'axis' of the input 'probability' array. The conditional
        probability is the probability of an event occurring given that another
        event has occurred. If the denominator is zero along the specified
        'axis', the result is set to zero to avoid division by zero errors.

        Parameters
        ----------
        probability : np.ndarray
            Array with probabilities
        axis : int
            Axis for which the conditional probability has to be calculated

        Returns
        -------
        np.ndarray
            Conditional probability
        """
        # Calculate the denominator as the sum of probabilities along the
        # specified 'axis'
        denominator = np.sum(probability, axis=axis)

        # Create a shape with all ones, except for the axis that will be used
        # for broadcasting
        shape = list(denominator.shape)
        shape.insert(axis, -1)

        # Calculate the conditional probability using element-wise division,
        # handling division by zero
        cond = np.divide(
            probability,
            denominator.reshape(shape),
            out=np.zeros_like(probability),
            where=(denominator != 0).reshape(shape),
        )

        return cond

    @staticmethod
    def calculate_boundaries(levels: np.ndarray):
        """
        Calculate the boundaries for the input array.

        This function calculates the boundaries between adjacent elements in
        the input 'arr' array. The boundaries are computed as the midpoints
        between consecutive elements, with additional values added at the start
        and end based on the step differences.

        Parameters
        ----------
        levels : np.ndarray
            Input array for which boundaries are to be calculated.

        Returns
        -------
        ndarray
            Array containing the calculated boundaries.
        """
        # Calculate the step differences between consecutive elements
        lower_step = levels[2] - levels[1]
        upper_step = levels[-1] - levels[-2]

        # Calculate midpoints between consecutive elements
        mid = (levels[1:] + levels[:-1]) / 2

        # Calculate the boundaries array by concatenating the midpoints with
        # additional boundary values
        bounds = np.concatenate([[mid[0] - lower_step], mid, [mid[-1] + upper_step]])

        return bounds



import numpy as np

from scipy.stats import norm

from ..statistics import Statistics
from ..stochastics.barrier.barrier_easternscheldt import BarrierEasternScheldt
from ..stochastics.discrete_probability import DiscreteProbability
from ..stochastics.model_uncertainty import ModelUncertainty
from ..stochastics.sea_level.sea_level_point import SeaLevelPoint
from ..stochastics.sigma_function import SigmaFunction
from ..stochastics.wind_speed import WindSpeed
from ....settings.settings import Settings
from .....common.interpolate import Interpolate
from .....common.probability import ProbabilityFunctions


class StatisticsEasternScheldt(Statistics):
    """
    Statistics class for the Eastern Scheldt
    Water systems: Eastern Scheldt
    """

    def __init__(self, settings: Settings):
        """
        Init the Statistics class for the Eastern Scheldt

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit initialisation method from parent
        super().__init__(settings)

        # Sea level
        self.sea_level = SeaLevelPoint(settings)

        # Storm surge duration
        self.storm_surge_duration = DiscreteProbability(
            settings.storm_surge_duration_probability
        )

        # Fase differences (between surge and tide)
        self.phase_surge_tide = DiscreteProbability(
            settings.phase_surge_tide_probability
        )

        # Sigma function
        self.sigma_function = SigmaFunction(settings)

        # Wind
        self.wind_direction = DiscreteProbability(settings.wind_direction_probability)
        self.wind_speed = WindSpeed(settings)
        self.wind_speed.correct_with_sigma_function(
            self.sigma_function, self.wind_direction
        )

        # Calculate P(u, m, r)
        self.__calculate_combined_probabilities()

        # Eastern Scheldt Barrier
        self.barrier = BarrierEasternScheldt(
            settings, self.wind_direction, self.wind_speed, self.sea_level
        )

        # Model uncertainty
        self.model_uncertainties = ModelUncertainty(settings)

        # Discrete, slow, fast stochatics
        self.stochastics_discrete = {
            "r": self.wind_direction.get_discretisation(),
            "k": [1, 2, 3, 4, 5, 6, 7, 8],
        }
        self.stochastics_fast = {
            "u": self.wind_speed.get_discretisation(),
            "m": self.sea_level.get_discretisation(),
            "d": self.storm_surge_duration.get_discretisation(),
            "p": self.phase_surge_tide.get_discretisation(),
        }
        self.stochastics_slow = {}

    def calculate_probability(
        self, wind_direction: float, closing_situation: int = 1, given: list = []
    ):
        """
        Calculate the probability of occurence for the discretisation given the wind direction.

        Parameters
        ----------
        wind_direction : float
            Wind direction
        closing_situation : int
            Closing situation, (irrelevant for Coast)
        given : list
            Given stochasts
        """
        # Probability of wind direction
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
        pwr = 1.0 if "r" in given else self.wind_direction.get_probability()[ir]

        # Probability of closing given a wind direction
        p_k = (
            1.0
            if "k" in given
            else self.barrier.calculate_closing_probability(
                wind_direction, closing_situation
            )
        )

        # Probability of a sea level and wind speed given a wind direction
        p_um_r = self.p_mur[:, :, ir]

        # If given, calculate the conditional probabilities
        if "u" in given:
            p_um_r[:] = ProbabilityFunctions.conditional_probability(p_um_r, axis=0)
        if "m" in given:
            p_um_r[:] = ProbabilityFunctions.conditional_probability(p_um_r, axis=1)

        # Combine all probabilities
        probability = (
            p_um_r[:, :, None, None]
            * self.storm_surge_duration.get_probability()[None, None, :, None]
            * self.phase_surge_tide.get_probability()[None, None, None, :]
            * p_k
            * pwr
        )

        # Return probability
        return probability

    def __calculate_combined_probabilities(self):
        # Statistics
        m = self.sea_level
        s = self.sigma_function
        r = self.wind_direction
        u = self.wind_speed

        # Initialize an empty matrix
        self.p_mur = np.zeros((len(u), len(m), len(r)))

        # Per wind direction
        for ir in range(len(r)):
            # Calculate the probability density of the sea level given the wind direction
            pd_m = ProbabilityFunctions.probability_density(
                m.get_discretisation(), m.get_exceedance_probability()[:, ir]
            )

            # If there is correlation (sigma > 0)
            if s.correlation[ir]:
                # Calculate sigma
                sigma = Interpolate.inextrp1d(
                    x=m.get_discretisation(), xp=s.sigma_sea_level, fp=s.sigma[:, ir]
                )
                if np.min(sigma) < 0.0:
                    raise ValueError()

                # Exceedance probability of the wind speed given the sea water level epm_r[Nwind, Nswl]
                snorm = (
                    self.wind_speed.k_u[:, ir][:, None] - m.epm_exp[:, ir][None, :]
                ) / sigma[None, :]
                ovkansen = 1 - norm.cdf(snorm)

                # Per sea level
                for im in range(len(m)):
                    pd_u = ProbabilityFunctions.probability_density(
                        u.get_discretisation(), ovkansen[:, im]
                    )
                    self.p_mur[:, im, ir] = pd_u.probability * pd_m.probability[im]

            # If there is no correlation (sigma <= 0)
            else:
                pd_u = ProbabilityFunctions.probability_density(
                    u.get_discretisation(), u.get_exceedance_probability()[:, ir]
                )
                self.p_mur[:, :, ir] = (
                    pd_m.probability[None, :] * pd_u.probability[:, None]
                )



import fiona as fn
import numpy as np
import os
import sqlite3

from shapely.geometry import Point, shape

from ...common.common import CommonFunctions
from ...common.enum import WaterSystem
from ...io.database_settings import DatabaseSettings


class Settings:
    # Non Hydra-NL related
    watersystem = None

    # Hydra-NL settings (Hydra-NL naming convention in comments)
    # Location-specific settings
    location = None  # LOCATIE
    x_coordinate = None  # XCOORDINAAT
    y_coordinate = None  # YCOORDINAAT
    region = None  # REGIO
    river = None  # RIVIER
    sea = None  # ZEE
    lake = None  # MEER

    # Block duration
    block_duration = None  # BLOKDUUR
    periods_block_duration = None  # NPERIODENWHJ

    # Wind speed
    u_repair = None  # REPAREER_U
    u_max = None  # UMAX
    u_step = None  # USTAP

    # Discharge
    q_repair = None  # REPAREER_Q
    q_min = None  # QMIN
    q_max = None  # QMAX
    q_step = None  # QSTAP (UNUSED IN PYDRA)
    q_limit = None  # QAFTOP

    # Sea level
    m_repair = None  # REPAREER_M
    m_min = None  # MMIN
    m_max = None  # MMAX
    m_step = None  # MSTAP
    sea_level_rise = None  # ZWS_STIJGING

    # Lake level
    a_repair = None  # REPAREER_A
    a_min = None  # MMIN
    a_max = None  # MMAX
    a_step = None  # MSTAP
    a_limit = None  # MAFTOP
    lake_level_rise = None  # MEERPEIL_STIJGING

    # Europoort
    europoort_barrier = None  # KERING
    europoort_barrier_distribution = None  # VERDELING
    europoort_barrier_mu = None  # MU
    europoort_barrier_sigma = None  # SIGMA
    europoort_barrier_alfa = None  # ALFA

    # Wave form (discharge (q) and lake level (a))
    base_duration = None  # BASISDUUR
    periods_base_duration = None  # NGEGEVENSBLOK
    waveshape_pw = None  # PW
    waveshape_time_step = None  # DISCRSTAPAFVMP
    top_duration_q = None  # TDQNM
    ifh_q = None  # IFHQ
    ifb_q = None  # IFBQ
    top_duration_a = None  # TDMNM
    ifh_a = None  # IFHM
    ifb_a = None  # IFBM

    # Lakes
    transition_lake_wave_shape = None  # VERSCHUIVING
    sigma_aq = None  # SIGMA_MQ

    # Lower riviers (and VZM)
    transitional_wind = None  # TRANSWIND
    fu = None  # FU

    # IJssel-Vechtdelta
    failure_probability_ramspol = None  # FAALKANSKERING

    # Model uncertainty
    model_uncertainty_water_level_steps = 7  # WS_ONZ_AANTAL
    model_uncertainty_wave_height_steps = 5  # GH_ONZ_AANTAL
    model_uncertainty_wave_period_steps = 5  # GP_ONZ_AANTAL

    # File paths
    # Database
    database_path = None  # DBRVW

    # Wind
    wind_direction_probability = None  # KANSRNM
    wind_speed_probability = None  # OVKANSU

    # Discharge
    discharge_probability = None  # OVKANSQ

    # Lake level
    lake_level_probability = None  # OVKANSM

    # Barrier Eastern Scheldt
    barrier_scenario_probability = None  # KANSENFAALSCENARIO
    barrier_closing_probability = None  # KANSSLUITINGEN

    # Phase difference between storm surge and tide
    phase_surge_tide_probability = None  # KANSENFASEVERSCHIL

    # Sea level
    sea_level_probability = None  # OVKANSM
    sea_level_probability_point1 = None  # OVKANSM_HOEKPUNT1
    sea_level_probability_point2 = None  # OVKANSM_HOEKPUNT2
    sea_level_probability_point3 = None  # OVKANSM_HOEKPUNT3

    # Sigma function
    sigma_function = None  # SIGMAFUNCTIE

    # Storm surge duration
    storm_surge_duration_probability = None  # KANSENSTORMDUUR

    # Wind-Sea level probability (for lower rivers/VZM)
    wind_sea_level_probability = None  # PWINDWESTNM

    def __init__(self, hrdlocation: str, database_path: str):
        """
        Init a Settings object for a location and database

        Parameters
        ----------
        hrdlocation : str
            The HRDLocation
        database_path : str
            Path to the database
        """
        # Save info
        self.location = hrdlocation
        self.database_path = database_path

        # Obtain the x, y coordinates and the water system
        con = sqlite3.connect(database_path)
        self.x_coordinate = int(
            con.execute(
                f"SELECT XCoordinate FROM HRDLocations WHERE Name = '{hrdlocation}'"
            ).fetchone()[0]
        )
        self.y_coordinate = int(
            con.execute(
                f"SELECT YCoordinate FROM HRDLocations WHERE Name = '{hrdlocation}'"
            ).fetchone()[0]
        )
        self.watersystem = WaterSystem(
            con.execute("SELECT GeneralId FROM General").fetchone()[0]
        )
        con.close()

        # Watersystem specific settings
        if CommonFunctions.is_lower_rivier(self.watersystem):
            self.__determine_lower_river_settings()

        # Location specific settings
        if CommonFunctions.is_coast(self.watersystem):
            self.__determine_sea_level_statistics_points()

        # Obtain and set settings
        with DatabaseSettings() as database:
            settings = database.get_settings(self.watersystem)

        # Loop through settings
        for _, row in settings.iterrows():
            # Lowercase
            key = row["SettingName"]
            if isinstance(key, str):
                key = key.lower()

            # Set settings
            if hasattr(self, key):
                setattr(self, key, row["SettingValue"])
            else:
                raise ValueError(f"[ERROR] Key '{key}' not defined in Settings.")

        # Collect all settings with statistical uncertainty (with '_metOnzHeid')
        self.__vars_with_statistical_uncertainty = {}
        for variable in [
            var
            for var in dir(self)
            if not callable(getattr(self, var)) and not var.startswith("__")
        ]:
            value = getattr(self, variable)
            if "_metOnzHeid" in str(value):
                self.__vars_with_statistical_uncertainty[variable] = value

    def include_statistical_uncertainty(self, include: bool):
        """
        Include statistical uncertainty.
        By default statistical uncertainty is included.

        Parameters
        ----------
        include : bool
            Whether to include statistical uncertainty
        """
        # Loop through all variables with statistical uncertainty
        for setting in self.__vars_with_statistical_uncertainty.keys():
            value = self.__vars_with_statistical_uncertainty[setting]
            setattr(
                self, setting, value if include else value.replace("_metOnzHeid", "")
            )

    def __determine_sea_level_statistics_points(self):
        """
        The sea level statistics along the coast, Wadden Sea and Western
        Scheldt are based on triangulation interpolation. This function
        determines which stations should be used for this interpolation.
        """
        # Coast
        if self.watersystem in [
            WaterSystem.COAST_SOUTH,
            WaterSystem.COAST_CENTRAL,
            WaterSystem.COAST_NORTH,
        ]:
            # Obtain table
            with DatabaseSettings() as database:
                refpoints = database.get_sea_level_statistic_points(self.watersystem)
                subsystems = database.get_sea_level_sub_systems(self.watersystem)

            # Determine the subregion
            if CommonFunctions.position_from_line(
                "north",
                (self.x_coordinate, self.y_coordinate),
                [refpoints.loc["Den Oever"].values, refpoints.loc["Den Helder"].values],
            ):
                subregion = 1
            elif self.y_coordinate > refpoints.loc["IJmuiden"].values[1]:
                subregion = 2
            elif self.y_coordinate > refpoints.loc["Hoek van Holland"].values[1]:
                subregion = 3
            else:
                subregion = 4

            # Determine subsystem
            subsystem = subsystems.loc[subregion].to_numpy()

        # Wadden sea
        elif self.watersystem in [
            WaterSystem.WADDEN_SEA_WEST,
            WaterSystem.WADDEN_SEA_EAST,
        ]:
            # Obtain table
            with DatabaseSettings() as database:
                refpoints = database.get_sea_level_statistic_points(self.watersystem)
                subsystems = database.get_sea_level_sub_systems(self.watersystem)

            # Determine the subregion
            if CommonFunctions.position_from_line(
                "east",
                (self.x_coordinate, self.y_coordinate),
                [
                    refpoints.loc["Lauwersoog Haven"].values,
                    refpoints.loc["Huibertgat"].values,
                ],
            ):
                subregion = 1
            elif CommonFunctions.position_from_line(
                "north",
                (self.x_coordinate, self.y_coordinate),
                [
                    refpoints.loc["Lauwersoog Haven"].values,
                    refpoints.loc["West-Terschelling"].values,
                ],
            ):
                subregion = 2
            elif CommonFunctions.position_from_line(
                "east",
                (self.x_coordinate, self.y_coordinate),
                [
                    refpoints.loc["Harlingen"].values,
                    refpoints.loc["West-Terschelling"].values,
                ],
            ):
                subregion = 3
            elif CommonFunctions.position_from_line(
                "east",
                (self.x_coordinate, self.y_coordinate),
                [
                    refpoints.loc["Den Oever"].values,
                    refpoints.loc["West-Terschelling"].values,
                ],
            ):
                subregion = 4
            elif self.x_coordinate < refpoints.loc["Den Helder"].values[0]:
                subregion = 5
            else:
                subregion = 6

            # Determine subsystem
            subsystem = subsystems.loc[subregion].to_numpy()

        # Western scheldt
        elif self.watersystem in [WaterSystem.WESTERN_SCHELDT]:
            # Set sea level points
            subsystem = np.array(
                [
                    "Zeewaterstand/Vlissingen/CondPovVlissingen_12u_zichtjaar2017_metOnzHeid.txt",
                    "Zeewaterstand/Hansweert/CondPovHansweert_12u_zichtjaar2017_metOnzHeid.txt",
                    "Zeewaterstand/Vlissingen virtueel/CondPovVlissingen-Additional_12u_zichtjaar2017_metOnzHeid.txt",
                ]
            )

        # Not implemented
        else:
            raise NotImplementedError(
                f"[ERROR] Vertices for sea level statistics not implemented for {self.watersystem}."
            )

        # If the subsystems contains three times the same file, specify the parameter related to a point
        if len(list(set(subsystem))) == 1:
            self.sea_level_probability = subsystem[0]

        # Otherwise, set sea level triangulation points
        else:
            self.sea_level_probability_point1 = subsystem[0]
            self.sea_level_probability_point2 = subsystem[1]
            self.sea_level_probability_point3 = subsystem[2]

    def __determine_lower_river_settings(self):
        """
        Selecting the right settings for the lower rivers
        """
        # Read from shape (MSTAP, MU, SIGMA, ALFA, QSTAP_Maas, QSTAP_rijn)
        PATH = os.path.join(
            os.path.split(os.path.dirname(__file__))[0],
            "..",
            "data",
            "settings",
            "lower_river_settings.shp",
        )
        with fn.open(PATH, "r") as shp:
            # Define the point using Shapely's Point
            point = Point(self.x_coordinate, self.y_coordinate)

            # Iterate through each subregion in the Shapefile
            for feature in shp:
                # Get the geometry of the feature
                geometry = shape(feature["geometry"])

                # Check if the point is within the geometry
                if point.within(geometry):
                    # Read settings
                    self.m_step = feature["properties"]["m_step"]
                    self.europoort_barrier_mu = feature["properties"]["ep_mu"]
                    self.europoort_barrier_sigma = feature["properties"]["ep_sigma"]
                    self.europoort_barrier_alfa = feature["properties"]["ep_alpha"]
                    if self.watersystem in [
                        WaterSystem.RHINE_TIDAL,
                        WaterSystem.EUROPOORT,
                    ]:
                        self.q_step = feature["properties"]["qs_rhine"]
                    elif self.watersystem == WaterSystem.MEUSE_TIDAL:
                        self.q_step = feature["properties"]["qs_meuse"]
                    else:
                        raise ValueError(
                            f"[ERROR] No lower river settings for {self.watersystem}."
                        )

                    # Stop the loop
                    break

    def __repr__(self):
        """
        Formatting when a settings object is printed. This will show all
        relevant settings with their values in a list.
        """
        lst = []
        for item in dir(self):
            if not item.startswith("_"):
                value = getattr(self, item)
                if value is None:
                    continue
                if isinstance(value, (str, float, int, bool)):
                    lst.append(f"{item:30s} = {value}")

        return "\n".join(lst)



import numpy as np

from .sea_level import SeaLevel
from .....settings.settings import Settings
from ......common.interpolate import Interpolate
from ......common.probability import ProbabilityFunctions
from ......io.file_hydranl import FileHydraNL


class SeaLevelTriangular(SeaLevel):
    """
    Class to describe the sea (water) level statistics.
    Sea level statistics are conditional on the wind direction.
    """

    def __init__(self, settings: Settings, epsilon: float = 1.0e-15):
        """
        For Coast systems, we use triangular interpolation of the statistics
        between three points.

        Parameters
        ----------
        settings : Settings
            The Settings object
        epsilon : float
            Maximum exceedance probability to determine the transformation
            table, used to take correlation into account.
        """
        # Inherit
        super().__init__()

        #  Read the exceedance probability of the sea level for three stations (triangular interpolation)
        m1, epm1 = FileHydraNL.read_file_ncolumns(settings.sea_level_probability_point1)
        m2, epm2 = FileHydraNL.read_file_ncolumns(settings.sea_level_probability_point2)
        m3, epm3 = FileHydraNL.read_file_ncolumns(settings.sea_level_probability_point3)

        # Check whether the amount of wind directions for each point are equal
        if epm1.shape[1] != epm2.shape[1] or epm2.shape[1] != epm3.shape[1]:
            raise ValueError(
                "aantal windrichtingen per hoekpunt zijn niet aan elkaar gelijk"
            )

        # Obtain the x and y positions of the three stations
        x1, y1 = FileHydraNL.read_file_ncolumns_loc(
            settings.sea_level_probability_point1
        )
        x2, y2 = FileHydraNL.read_file_ncolumns_loc(
            settings.sea_level_probability_point2
        )
        x3, y3 = FileHydraNL.read_file_ncolumns_loc(
            settings.sea_level_probability_point3
        )

        # Check if the exceedance probability of the lowest sea level are all equal to 1
        if any(epm1[0, :] != 1) or any(epm2[0, :] != 1) or any(epm3[0, :] != 1):
            raise ValueError(
                "[ERROR] Exceedance probability of the lowest sea level is not equal to 1"
            )

        # Calculate the lowest sea level for the hr location
        m_min = Interpolate.triangular_interpolation(
            x1,
            y1,
            m1[0],
            x2,
            y2,
            m2[0],
            x3,
            y3,
            m3[0],
            settings.x_coordinate,
            settings.y_coordinate,
        )

        # Equidistant filling vector with seawater levels
        self.m = ProbabilityFunctions.get_hnl_disc_array(
            m_min, settings.m_max, settings.m_step
        )
        self.nm = len(self.m)

        # Number of wind directions
        nr = epm1.shape[1]

        # Create matrix
        self.epm = np.zeros([self.nm, nr])

        # Iterate over each wind direction
        for ir in range(0, nr):
            # For high sea water levels, the exceedance probability may be 0
            # Find the location where the highest sea level has the greatest exceedance probability no equal to 0.
            largest_ep = 0.0
            i = 1
            while largest_ep <= 0.0:
                corners = [(m1, epm1), (m2, epm2), (m3, epm3)]
                for _m, _epm in corners:
                    nlm = len(_m) - i
                    if _epm[nlm, ir] > largest_ep:
                        largest_ep = _epm[nlm, ir]
                        ir_epm = _epm[:nlm, ir]
                i += 1

            # Determine the sea level discretisation for the three stations
            ir_m1 = Interpolate.inextrp1d_log_probability(
                ir_epm, epm1[:, ir][::-1], m1[::-1]
            )
            ir_m2 = Interpolate.inextrp1d_log_probability(
                ir_epm, epm2[:, ir][::-1], m2[::-1]
            )
            ir_m3 = Interpolate.inextrp1d_log_probability(
                ir_epm, epm3[:, ir][::-1], m3[::-1]
            )

            # Use triangular interpolation to calculate the sea level at the specified exceedance probability at the given location.
            ir_m = Interpolate.triangular_interpolation(
                x1,
                y1,
                ir_m1,
                x2,
                y2,
                ir_m2,
                x3,
                y3,
                ir_m3,
                settings.x_coordinate,
                settings.y_coordinate,
            )

            # Als de triangulaire interpolatie wordt toegepast in het geval van extrapolatie
            # kan de vector _m een verloop hebben dat niet langer monotoon is. Dit wordt
            # gecorrigeerd door die stukjes van de vector _m weg te gooien.
            idx = np.diff(ir_m) > 0.0
            if not idx.all():
                pos = np.where(idx)[0]
                pos = np.unique(np.concatenate([pos, pos + 1]))
                ir_m = ir_m[pos]
                ir_epm = ir_epm[pos]

            # Interpoleer logaritmisch de conditionele overschrijdingskansen van de zeewaterstand gegeven de windrichting naar het gewenste rooster
            ir_epm_int = np.exp(Interpolate.inextrp1d(self.m, ir_m, np.log(ir_epm)))

            # Door extrapolatie kunnen overschrijdingskansen groter dan 1 ontstaan zijn. Deze worden gelijkgesteld aan 1.
            # Door extrapolatie kunnen overschrijdingskansen kleiner dan 0 ontstaan zijn. Deze worden gelijkgesteld aan 0.
            # Door extrapolatie kan de mimimale zeewaterstand toch een overschrijdingskans van kleiner dan 1
            # krijgen. Dit betreft dan een onbelangrijke richting. Hiervoor wordt gecorrigeerd dat de
            # overschrijdingskans van de laagste zeewaterstand toch gelijk is aan 1.0
            ir_epm_int[0] = 1.0
            ir_epm_int = np.clip(ir_epm_int, 0.0, 1.0)

            # Add the results to the exceedance probability matrix
            self.epm[:, ir] = ir_epm_int

        # Verwerk de zeespiegelstijging
        self.m = self.m + settings.sea_level_rise

        # Bereken de transformatietabel van de OVERschrijdingskansen naar exponentiële ruimte
        # voor het correlatiemodel van de zeewaterstand en de windsnelheid.
        # Zie ook vgl. 4.7 van [Geerse, 2012], merk op dat het daar ONDERschrijdingskansen betreft#
        self.epm_exp = -np.log(np.maximum(self.epm, epsilon))



from enum import Enum


class WaterSystem(Enum):
    """
    A class representing different water systems in the Netherlands.

    WaterSystem is an enumeration representing a specific water system in The
    Netherlands and is defined with a unique integer value (WBI2017 Appendix
    II). The enumeration members include both tidal and non-tidal water
    systems, such as rivers, lakes, seas, and coastal regions.
    """

    RHINE_NON_TIDAL = 1
    MEUSE_NON_TIDAL = 2
    RHINE_TIDAL = 3
    MEUSE_TIDAL = 4
    IJSSEL_DELTA = 5
    VECHT_DELTA = 6
    IJSSEL_LAKE = 7
    MARKER_LAKE = 8
    WADDEN_SEA_EAST = 9
    WADDEN_SEA_WEST = 10
    COAST_NORTH = 11
    COAST_CENTRAL = 12
    COAST_SOUTH = 13
    EASTERN_SCHELDT = 14
    WESTERN_SCHELDT = 15
    COAST_DUNES = 16
    EUROPOORT = 17
    MEUSE_VALLEY_NON_TIDAL = 18
    VELUWE_LAKES = 19
    GREVELINGEN = 20
    VOLKERAK_ZOOMMEER = 21
    HOLLANDSCHE_IJSSEL = 22
    DIEFDIJK = 23


class Breakwater(Enum):
    """
    Breakwater classes for profiles.

    The 'Breakwater' is an enumeration representing specific types of
    breakwaters located in front of the profile.
    """

    NO_BREAKWATER = 0
    CAISSON = 1
    VERTICAL_WALL = 2
    RUBBLE_MOUND = 3


class WaveShapeType(Enum):
    """
    Different wave form types.

    The 'WaveShapeType' is an enumeration representing specific types of wave
    forms used for slow stochastics. It can either be a discharge wave or a
    lake level wave.
    """

    DISCHARGE = 0
    LAKE_LEVEL = 1



import numpy as np

from scipy.stats import norm

from .discrete_probability import DiscreteProbability
from .sigma_function import SigmaFunction
from ....settings.settings import Settings
from .....common.interpolate import Interpolate
from .....io.file_hydranl import FileHydraNL


class WindSpeed:
    """
    Class to describe the wind speed statistics.
    Wind speed statistics are conditional on the wind direction.
    """

    def __init__(self, settings: Settings):
        """
        Constructor class for the WindSpeed statistics.

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Default None, only initialized when correcting with sigma function
        self.k_u = None

        # Read wind speed
        self.__read_wind_speed(settings)

        # Scale for block_duration
        if (settings.block_duration != 12.0) and (settings.block_duration is not None):
            self.epu = 1.0 - (1.0 - self.epu) ** (settings.block_duration / 12.0)
        self.block_duration_wind = settings.block_duration

    def __len__(self):
        """
        Return the number of wind speed discretisations

        Returns
        -------
        int
            Number of discretisations
        """
        return self.nu

    def __read_wind_speed(self, settings: Settings) -> None:
        """
        Read and process the wind speed statistics

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Read the wind speed and exceedance probabilities
        lu, eplu = FileHydraNL.read_file_ncolumns(settings.wind_speed_probability)

        # If the u_max in the settings is larger than the maximum wind speed in the statistics
        if settings.u_max > max(lu):
            # Calculate the number of wind speeds to add
            nuhoger = int(round(max(1.0, settings.u_max - max(lu))))

            # Add the wind speeds with a step of 1
            u = np.concatenate(
                [lu, np.linspace(max(lu), settings.u_max, nuhoger + 1)[1:]]
            )
            self.epu = np.zeros((len(u), eplu.shape[1]))

            # Add exceedance probabilities (extrapolate)
            for i, col in enumerate(eplu.T):
                # If the last value is 0 or less, add zeroes
                if col[-1] <= 0.0:
                    self.epu[len(lu) :, i] = 0.0

                # Otherwise, extrapolate logarithmically
                else:
                    self.epu[:, i] = np.exp(
                        Interpolate.inextrp1d(x=u, xp=lu, fp=np.log(col))
                    )

        # Otherwise, do nothing
        else:
            u = lu[:]
            self.epu = eplu[:, :]

        # If a stepsize for the windspeed is defined
        if settings.u_step is not None:
            # Create a new grid
            ugrid = np.concatenate(
                [np.arange(0, settings.u_max, settings.u_step), [settings.u_max]]
            )

            # Interpolate the exceedance probabilities for this grid
            epu = []
            for col in self.epu.T:
                logep = col
                logep[logep > 0.0] = np.log(logep[logep > 0.0])
                epu.append(np.exp(Interpolate.inextrp1d(x=ugrid, xp=u, fp=logep)))

            # Vervang overschrijdingskansen en u
            self.epu = np.c_[epu].T
            u = ugrid

        # Add the results to the object
        self.u = u
        self.nu = len(u)

    def correct_with_sigma_function(
        self, sigma_function: SigmaFunction, wind_direction: DiscreteProbability
    ) -> None:
        """
        Correct with statistics with sigma function.

        Parameters
        ----------
        sigma_function : SigmaFunction
            The SigmaFunction statistics object
        """
        # Init a matrix
        self.k_u = np.ones((self.nu, len(wind_direction))) * -99.0

        # Per windrichting
        for ir in range(len(wind_direction)):
            # Als er correlatie is (sigma > 0)
            if sigma_function.correlation[ir]:
                lim1 = -4 * max(sigma_function.sigma[:, ir])
                lim2 = 20.0 + 4 * max(sigma_function.sigma[:, ir])
                f_y_m = np.linspace(lim1, lim2, 351)
                f_y_sigma = self.__wind_transformation(
                    f_y_m,
                    0.0,
                    sigma_function.sigma_sea_level,
                    sigma_function.sigma[:, ir],
                )

                # Kansverdeling schalen of de hoogste ONDERschrijdingskans gelijk stellen aan 1
                if f_y_sigma[-1] > 1.0:
                    f_y_sigma /= f_y_sigma[-1]
                else:
                    f_y_sigma[-1] = 1.0

                ondkansu = 1.0 - self.epu[:, ir]
                self.k_u[:, ir] = Interpolate.inextrp1d(
                    x=ondkansu, xp=f_y_sigma, fp=f_y_m
                )

    def __wind_transformation(self, y, mu, transzee, sigmafunctie):
        """
        Transform the wind speed statistics for correlation with the sea level.
        """
        if min(sigmafunctie) <= 0.0:
            raise ValueError()

        stap = 0.05
        x = np.arange(0, 20 + 0.1 * stap, stap)
        sigma = Interpolate.inextrp1d(x=x, xp=transzee, fp=sigmafunctie)

        if min(sigma) <= 0.0:
            raise ValueError()

        yn = (y[None, :] - x[:, None] - mu) / sigma[:, None]

        dx = np.diff(np.r_[x[0], (x[1:] + x[:-1]) / 2.0, x[-1]])

        f_y_sigma = (
            np.exp(-x[:, None]) * norm.cdf(x=yn, loc=0.0, scale=1.0) * dx[:, None]
        ).sum(0)

        return f_y_sigma

    def get_discretisation(self) -> np.ndarray:
        """
        Return the wind speed discretisation.

        Returns
        -------
        np.ndarray
            1D array with discretisation
        """
        return self.u

    def get_exceedance_probability(self) -> np.ndarray:
        """
        Return exceedance probility of the wind speed, conditional on the wind
        direction.

        Returns
        -------
        np.ndarray
            A 2D array with the wind speed exceedance probability conditional
            on the wind direction (r x u)
        """
        return self.epu



import numpy as np


class Interpolate:
    """
    A class with common functions used for interpolation.
    """

    @staticmethod
    def inextrp1d(x, xp, fp):
        """
        Interpolate an array along the given axis.
        Similar to np.interp, but with extrapolation outside range.

        Parameters
        ----------
        x : np.array
            Array with positions to interpolate at
        xp : np.array
            Array with positions of known values
        fp : np.array
            Array with values as known positions to interpolate between

        Returns
        -------
        np.array
            interpolated array
        """
        # Determine lower bounds
        intidx = np.minimum(np.maximum(0, np.searchsorted(xp, x) - 1), len(xp) - 2)

        # Determine interpolation fractions
        fracs = (x - xp[intidx]) / (xp[intidx + 1] - xp[intidx])

        # Interpolate (1-frac) * f_low + frac * f_up
        f = (1 - fracs) * fp[intidx] + fp[intidx + 1] * fracs

        return f

    @staticmethod
    def inextrp1d_log_probability(
        x: np.ndarray, xp: np.ndarray, fp: np.ndarray
    ) -> np.ndarray:
        """
        Perform 1D log-interpolation of log-probability values using piecewise
        linear interpolation.

        This function interpolates log-probability values for input points 'x'
        using a given set of log-probabilities 'fp' and corresponding
        probabilities 'xp'. The interpolation is performed in a piecewise
        manner, taking into account the logarithmic nature of the data.

        Parameters
        ----------
        x : np.ndarray
            Input points at which log-probabilities are to be interpolated.
        xp : np.ndarray
            Known data points representing the probabilities.
        fp : np.ndarray
            Log-probabilities corresponding to the known data points 'xp'.

        Returns
        -------
        np.ndarray
            Interpolated log-probability values for the input points 'x'.

        Raises:
            ValueError: If the provided xp values (probabilities) are not
            monotonically increasing.

        Notes
        -----
        -   The input array 'xp' should be strictly increasing; otherwise, a
            ValueError is raised.
        """
        # Check whether the probabilities are monotonically decreasing
        if not (np.diff(xp) >= 0.0).all():
            raise ValueError(
                "[ERROR] xp-values are not monotonically increasing or decreasing"
            )

        # Determine lower bounds
        intidx = np.minimum(np.maximum(0, np.searchsorted(xp, x) - 1), len(xp) - 2)

        # Split in log and linear part
        iszero = xp[intidx] == 0.0
        intidx_log = intidx[~iszero]
        intidx_lin = intidx[iszero]

        # Determine interpolation fractions
        fracs = np.zeros(len(x), dtype=float)
        fracs[~iszero] = (np.log(x[~iszero]) - np.log(xp[intidx_log])) / (
            np.log(xp[intidx_log + 1]) - np.log(xp[intidx_log])
        )
        fracs[iszero] = (x[iszero] - xp[intidx_lin]) / (
            xp[intidx_lin + 1] - xp[intidx_lin]
        )

        # Interpolate (1-frac) * f_low + frac * f_up
        f = (1 - fracs) * fp[intidx] + fp[intidx + 1] * fracs

        return f

    @staticmethod
    def triangular_interpolation(
        x1: float,
        y1: float,
        h1: float,
        x2: float,
        y2: float,
        h2: float,
        x3: float,
        y3: float,
        h3: float,
        x: float,
        y: float,
    ) -> float:
        """
        Performs triangular interpolation to calculate the parameter value
        (water level) at a given point within a triangle.

        Parameters
        ----------
        x1 : float
            X-coordinate of the first input location.
        y1 : float
            Y-coordinate of the first input location.
        h1 : float
            Parameter value (water level) at the first input location.
        x2 : float
            X-coordinate of the second input location.
        y2 : float
            Y-coordinate of the second input location.
        h2 : float
            Parameter value (water level) at the second input location.
        x3 : float
            X-coordinate of the third input location.
        y3 : float
            Y-coordinate of the third input location.
        h3 : float
            Parameter value (water level) at the third input location.
        x : float
            X-coordinate of the point to interpolate.
        y : float
            Y-coordinate of the point to interpolate.

        Returns
        -------
        float
            The interpolated parameter value (water level) at the given point.
        """
        # Check if the input locations are invalid
        if (
            (x1 == x2 and y1 == y2)
            or (x1 == x3 and y1 == y3)
            or (x2 == x3 and y2 == y3)
        ):
            raise ValueError(
                "[ERROR] Two input locations in triangle interpolation are exactly the same. This is not allowed."
            )
        if (x1 == x2 == x3) or (y1 == y2 == y3):
            raise ValueError(
                "[ERROR] The three input locations in triangle interpolation are collinear. This is not allowed."
            )

        # Check if the locations are on the same line
        a_1 = (y1 - y2) / (x1 - x2)
        b_1 = y1 - a_1 * x1
        a_2 = (y3 - y2) / (x3 - x2)
        b_2 = y3 - a_1 * x3
        if (a_1 == a_2) and (b_1 == b_2):
            raise ValueError(
                "[ERROR] The three input locations in triangle interpolation are collinear. This is not allowed."
            )

        # Calculate auxiliary parameter a
        a = x1 * (y2 - y3) - x2 * (y1 - y3) + x3 * (y1 - y2)

        # Parameter a should not be equal to 0
        if a == 0.0:
            raise ValueError(
                "[ERROR] In triangle interpolation, there is an auxiliary parameter (a) that should not be equal to 0, but it is currently set to 0."
            )

        # Calculate parameters B1, B2 en B3
        b1 = h1 * (y2 - y3) - h2 * (y1 - y3) + h3 * (y1 - y2)
        b2 = x1 * (h2 - h3) - x2 * (h1 - h3) + x3 * (h1 - h2)
        b3 = (
            x1 * (y2 * h3 - y3 * h2)
            - x2 * (y1 * h3 - y3 * h1)
            + x3 * (y1 * h2 - y2 * h1)
        )

        # Calculate the value at (x, y) using the triangular interpolation
        triangular_interp = (b1 / a) * x + (b2 / a) * y + b3 / a
        return triangular_interp


class InterpStruct:
    """
    Interpolation helper class.

    This class provides functionality for 1D interpolation of arrays along a
    specified axis. It is designed to interpolate arrays using piecewise linear
    interpolation based on provided data points.
    """

    def __init__(self, x: np.ndarray, xp: np.ndarray):
        """
        Initialize the InterpStruct object and calculate interpolation factors.

        This method is used internally to calculate interpolation factors based
        on given data points 'x' and 'xp'. It sets the 'xp', 'x', 'intidx', and
        'fracs' attributes of the InterpStruct object.

        Parameters
        ----------
        x : np.ndarray
            Input data points used for interpolation.
        xp : np.ndarray
            Known data points representing the interpolation axis.
        """
        self.xp = np.asarray(xp)
        self.x = np.asarray(x)
        self.intidx = np.minimum(
            np.maximum(0, np.searchsorted(self.xp, self.x) - 1), len(self.xp) - 2
        )
        self.fracs = (self.x - self.xp[self.intidx]) / (
            self.xp[self.intidx + 1] - self.xp[self.intidx]
        )

    def interp(
        self,
        fp: np.ndarray,
        axis: int = 0,
        extrapolate: bool = True,
        left: bool = None,
        right: bool = None,
    ):
        """
        Interpolate a (multidimensional) array along the given axis using
        piecewise linear interpolation.

        Parameters
        ----------
        fp : np.ndarray
            Array with values to interpolate.
        axis : int, optional
            Axis along which the interpolation is performed (default is 0).
        extrapolate : bool, optional
            If True, allow extrapolation of values outside the range of 'xp'.
            If False, set the values outside the interpolation range to NaN
            (default is True).
        left : float or None, optional
            Value to use for extrapolation on the left side. If None, the left
            extrapolation will result in NaN (default is None).
        right : float or None, optional:
            Value to use for extrapolation on the right side. If None, the
            right extrapolation will result in NaN (default is None).

        Returns
        -------
        ndarray
            Interpolated multidimensional array.

        Raises
        ------
            ValueError: If the given axis is higher than the dimensions of the
            'fp' array.
        """
        # Convert to array if needed
        if isinstance(fp, list):
            fp = np.asarray(fp)

        # Check given axis and fp shape
        if axis > fp.ndim - 1:
            raise ValueError(
                f"Given axis ({axis}) is higher than dimensions of fp array ({fp.ndim})."
            )

        # Create shape with all ones, except for the array which is used for
        # interpolation
        shape = [1] * fp.ndim
        shape[axis] = len(self.fracs)

        # Interpolate (1-frac) * f_low + frac * f_up
        f = (1 - self.fracs.reshape(shape)) * np.take(
            fp, self.intidx, axis=axis
        ) + np.take(fp, self.intidx + 1, axis=axis) * self.fracs.reshape(shape)

        # If 'extrapolate' is False, set values outside the interpolation range
        # to NaN or the specified values
        if not extrapolate:
            if left is not None:
                f[(self.x < self.xp[0])] = left
            else:
                f[(self.x < self.xp[0])] = np.nan
            if right is not None:
                f[(self.x > self.xp[-1])] = right
            else:
                f[(self.x > self.xp[-1])] = np.nan

        return f

    def interp_angle(
        self, fp: np.ndarray, axis: int = 0, extrapolate: bool = True
    ) -> np.ndarray:
        """
        Interpolate angles (in degrees) using piecewise linear interpolation.

        This function performs interpolation of angles represented by 'fp'
        using piecewise linear interpolation. The angles are given in degrees
        and can be provided as a 1D or 2D array.

        Parameters
        ----------
        fp : np.ndarray
            Input angles (in degrees) to be interpolated.
        axis : int, optional
            The axis along which the interpolation is performed (default is 0).
        extrapolate : bool, optional
            If True, allow extrapolation of values outside the range of
            'self.xp'. If False, set the angles outside the interpolation
            range to NaN (default is True).

        Returns
        -------
        np.ndarray
            Interpolated angles in degrees.
        """
        # Convert angles from degrees to cosine and sine values and perform
        # interpolation along the specified 'axis'
        xint = self.interp(np.cos(np.radians(fp)), axis=axis)
        yint = self.interp(np.sin(np.radians(fp)), axis=axis)

        # Compute the interpolated angle in degrees using arctan2 and wrap it
        # to the range [0, 360)
        f = np.degrees(np.arctan2(yint, xint)) % 360

        # If 'extrapolate' is False, set angles outside the interpolation range
        # to NaN
        if not extrapolate:
            f[(self.x < self.xp[0]) | (self.x > self.xp[-1])] = np.nan

        return f



from .statistics import Statistics
from .water_systems.statistics_coast import StatisticsCoast
from .water_systems.statistics_eastern_scheldt import StatisticsEasternScheldt
from .water_systems.statistics_ijssel_vechtdelta import StatisticsIJsselVechtdelta
from .water_systems.statistics_lake import StatisticsLake
from .water_systems.statistics_lower_river import StatisticsLowerRiver
from .water_systems.statistics_upper_river import StatisticsUpperRiver
from ...settings.settings import Settings
from ....common.enum import WaterSystem


class StatisticsFactory:
    """
    A factory class to generate the right Statistics object for a given Settings object.

    Attributes
    ----------
    WATER_SYSTEM_STATISTICS_MAP : dict
        A dictionary containing the corresponding Statistics class for a WaterSystem
    """

    # Dictionary with Statistic classes for each WaterSystem
    WATER_SYSTEM_STATISTICS_MAP = {
        # Upper River
        WaterSystem.RHINE_NON_TIDAL: StatisticsUpperRiver,
        WaterSystem.MEUSE_NON_TIDAL: StatisticsUpperRiver,
        WaterSystem.MEUSE_VALLEY_NON_TIDAL: StatisticsUpperRiver,
        # Lower River
        WaterSystem.RHINE_TIDAL: StatisticsLowerRiver,
        WaterSystem.MEUSE_TIDAL: StatisticsLowerRiver,
        # TODO: WaterSystem.EUROPOORT
        # Coast
        WaterSystem.WADDEN_SEA_EAST: StatisticsCoast,
        WaterSystem.WADDEN_SEA_WEST: StatisticsCoast,
        WaterSystem.COAST_NORTH: StatisticsCoast,
        WaterSystem.COAST_CENTRAL: StatisticsCoast,
        WaterSystem.COAST_SOUTH: StatisticsCoast,
        WaterSystem.WESTERN_SCHELDT: StatisticsCoast,
        # Eastern Scheldt
        WaterSystem.EASTERN_SCHELDT: StatisticsEasternScheldt,
        # Lakes
        WaterSystem.IJSSEL_LAKE: StatisticsLake,
        WaterSystem.MARKER_LAKE: StatisticsLake,
        # TODO: WaterSystem.VELUWE_LAKES
        # TODO: WaterSystem.GREVELINGEN
        # IJssel-Vecht Delta
        WaterSystem.VECHT_DELTA: StatisticsIJsselVechtdelta,
        WaterSystem.IJSSEL_DELTA: StatisticsIJsselVechtdelta,
        # Volkerak-Zoommeer
        # TODO: WaterSystem.VOLKERAK_ZOOMMEER
        # Hollandsche IJssel
        # TODO: WaterSystem.HOLLANDSCHE_IJSSEL
        # Other
        # WaterSystem.COAST_DUNES
        # WaterSystem.DIEFDIJK
    }

    @staticmethod
    def get_statistics(settings: Settings) -> Statistics:
        """
        Returns the Statistics object for the specified water system.

        Parameters
        ----------
        settings : Settings
            A Settings object to get statistics for.

        Returns
        -------
        Statistics
            A Statistics object for the specified water system.

        Raises
        ------
        NotImplementedError
            If statistics for the specified water system are not implemented.
        """
        # Obtain the right Statistics class
        statistics_class = StatisticsFactory.WATER_SYSTEM_STATISTICS_MAP.get(
            settings.watersystem
        )

        # Return if the class is found, otherwise raise an error
        if statistics_class:
            return statistics_class(settings)
        else:
            raise NotImplementedError(
                f"[ERROR] Statistics for '{settings.watersystem.name}' not implemented (ID: {settings.watersystem.value})"
            )



import numpy as np

from .wave_shape import WaveShape
from ....settings.settings import Settings
from .....common.enum import WaterSystem, WaveShapeType
from .....common.interpolate import Interpolate
from .....io.file_hydranl import FileHydraNL


class LakeLevel:
    """
    Class to describe the lake (water) level statistics.
    """

    def __init__(self, settings: Settings):
        """
        Constructor class for LakeLevel statistics.

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Read exceedance probability peak lake level
        apeak, epapeak = FileHydraNL.read_file_2columns(settings.lake_level_probability)

        # Adjust for lake level rise
        apeak += settings.lake_level_rise

        # Lowest lake level
        a_min_piek = max(apeak[0] - settings.lake_level_rise, settings.a_min)

        # Creeer rooster met meerpeilen voor de trapezia
        self.apeak = np.r_[
            np.arange(a_min_piek, settings.a_max, settings.a_step), settings.a_max
        ]
        self.napeak = len(self.apeak)

        # Als de meerpeilen bekend zijn kunnen de trapezia worden geïnitialiseerd
        # (Vul de vector met meerpeilen)
        self.wave_shape = WaveShape(settings, type=WaveShapeType.LAKE_LEVEL)
        self.wave_shape.initialise_wave_shapes(
            self.apeak, climate_change=settings.lake_level_rise
        )

        # Creeer rooster met meerpeilen voor het opdelen van de golfvormen in blokken
        # TODO self.invoergeg.m_min #= MLAAGST? Moet dit dan kol1.min() zijn?
        # self.mblok = roosterblokken(self.invoergeg.m_min, self.mpiek)
        self.ablok = self.apeak.copy()
        self.nablok = len(self.ablok)

        # Blaas de tabel met overschrijdingskansen voor het meerpeil op
        # De interpolatie wordt logaritmische uitgevoerd.
        self.epapeak = np.exp(
            Interpolate.inextrp1d(x=self.apeak, xp=apeak, fp=np.log(epapeak))
        )
        self.epapeak[self.epapeak > 1] = 1.0

        # IJssel-Vechtdelta and VZM related
        if settings.watersystem in [
            WaterSystem.IJSSEL_DELTA,
            WaterSystem.VECHT_DELTA,
            WaterSystem.VOLKERAK_ZOOMMEER,
        ]:
            # Lokale paramters
            N = 376

            # Geef verwachtingswaarde en standaarddeviatie van de normale verdeling in de getransformeerde ruimte een waarde
            mu = 0.0
            sigma = settings.sigma_aq  # invoergeg.sigma_mq(ig)

            # Zet de gegevens voor de verfijnde interpolatietabel (de y-waarden, ofwel de K_sigma(s) uit vgl. 4.7 van [Geerse, 2003.129x]
            # Bereken voor elke y-waarde de integraal 4.3 [Geerse, 2003]
            f_y_k = np.linspace(-2 - 4 * sigma, 20 + 4 * sigma, N)

            # Verfijnde tabel klaarzetten van de cumulatieve verdelingsfunctie onder transformatie van s naar y
            # Zie ook vgl. 4.3 van [Geerse, 2003]

            # Maak de vector met y-waarden:
            # Bereken voor elke y-waarde de integraal 4.3 [Geerse, 2003]
            f_y_sigma = self.wave_shape.cum_norm_s_naar_y(f_y_k, mu, sigma)

            # Kansverdeling schalen of de hoogste ONDERschrijdingskans gelijk stellen aan 1
            if f_y_sigma[-1] > 1.0:
                f_y_sigma /= f_y_sigma[-1]
            else:
                f_y_sigma[-1] = 1.0

            for i in range(N):
                if f_y_sigma[i] > 0:
                    D = i
                    break

            # Een ONDERschrijdingskans gelijk aan 0.0 mag worden meegenomen.
            if D > 0:
                D -= 1

            self.k_apeak = Interpolate.inextrp1d(
                x=1 - self.epapeak, xp=f_y_sigma[D:], fp=f_y_k[D:]
            )

        #  Bereken de momentane overschrijdingskansen van het meerpeil
        self.mom_ovkans_a = self.wave_shape.instantaneous_exceedance_probability(
            self.epapeak, self.ablok
        )

        # Transition
        if (settings.transition_lake_wave_shape is not None) and (
            settings.transition_lake_wave_shape != 0.0
        ):
            self.wave_shape.transition_wave(settings.transition_lake_wave_shape)

        # Filter overschrijdingskansen die kleiner dan 0 of grote dan 1 zijn. Dit kan voorkomen door floating point precision
        self.upablok = np.clip(1.0 - self.mom_ovkans_a, 0.0, 1.0)

    def __len__(self):
        """
        Return the number of discharge discretisations.

        Returns
        -------
        int
            Number of discretisations
        """
        return self.nablok

    def get_discretisation(self) -> np.ndarray:
        """
        Return the discharge discretisation.

        Returns
        -------
        np.ndarray
            1D array with discretisation
        """
        return self.ablok

    def get_exceedance_probability(self) -> np.ndarray:
        """
        Return exceedance probility of the discharge

        Returns
        -------
        np.ndarray
            A 1D array with the discharge exceedance probability
        """
        return self.upablok

    def get_wave_shape(self) -> WaveShape:
        """
        Return the wave form statistics object.

        Returns
        -------
        WaveShape
            The WaveShape statistics
        """
        return self.wave_shape



import numpy as np

from abc import ABC, abstractmethod

from ..discrete_probability import DiscreteProbability
from .....settings.settings import Settings


class Barrier(ABC):
    def __init__(self, settings: Settings, wind_direction: DiscreteProbability):
        # Save key information
        self.settings = settings
        self.wind_direction = wind_direction
        self.k = None
        self.nk = None

    def __len__(self):
        """
        Return the number of barrier discretisations.

        Returns
        -------
        int
            Number of discretisations
        """
        return self.nk

    def get_discretisation(self) -> np.ndarray:
        """
        Return the barrier discretisation.

        Returns
        -------
        np.ndarray
            1D array with discretisation
        """
        return self.k

    @abstractmethod
    def calculate_closing_probability(
        self, wind_direction: float, closing_situation: int
    ) -> np.ndarray:
        pass



import numpy as np

from .barrier import Barrier
from ..discrete_probability import DiscreteProbability
from ..sea_level.sea_level import SeaLevel
from ..wind_speed import WindSpeed
from ....loading.loading_model.loading_model import LoadingModel
from .....settings.settings import Settings
from ......common.interpolate import InterpStruct, Interpolate
from ......io.database_hr import DatabaseHR
from ......io.file_hydranl import FileHydraNL


class BarrierEasternScheldt(Barrier):
    def __init__(
        self,
        settings: Settings,
        wind_direction: DiscreteProbability,
        wind_speed: WindSpeed,
        sea_level: SeaLevel,
    ):
        # Inherit
        super().__init__(settings, wind_direction)

        # Lees kansen op sluitscenario's: de kans op bemand of onbemand
        self.t = np.array([0.0, 1.0])

        # pfk bevat de kans op een aantal falende schuiven (bijv 0, 16, 31, 62) voor
        # een bemande dan wel onbemande sluiting, gegeven dat de kering faalt dus
        self.k, self.p_fail_k = FileHydraNL.read_file_ncolumns(
            settings.barrier_scenario_probability
        )
        self.k = self.k.astype(int)

        # Lees kansen op sluitscenario's: de kans op sluiting gegeven de zeewaterstand en aantal falende sluizen
        self.n_closing = len(self.k)
        self.nk = len(self.t) * len(self.k)

        # sluitkansenk bevat de kans op een sluitvraag (al dan niet succesvol) gegeven de zeewaterstand
        # Voor een stategiesluiting, handmatige sluiting, noodsluiting door het systeem, of geen sluiting
        self.m_closing, self.p_closing_k = FileHydraNL.read_file_ncolumns(
            settings.barrier_closing_probability
        )

        # Database connection
        with DatabaseHR(settings.database_path) as database:
            # Read closing situations
            self.closing_situations = database.get_closing_situations_eastern_scheldt()

            # Maak sluitpeilen structure (u, m, r, d, p)
            self.closing_levels = LoadingModel(
                None, None, ["u", "m", "r", "d", "p"], ["h_rpb"]
            )
            self.closing_levels.initialise(
                database.get_closing_levels_table_eastern_scheldt()
            )

        # h_roompot_buiten bevat de waterstand bij RPBU. Het sluitpeil van de oosterscheldekering is 3.0 m+NAP
        # bij hogere waterstanden zal de kering dus (pogen) te sluiten. Nu is de kans op al dan niet sluiten
        # ook afhankelijk van deze waterstand. Hoe hoger de waterstand, hoe groter de kans dat het misgaat.
        # Daarom zit er in de database een tabel met optredende waterstanden bij Roompot Buiten. Wanneer
        # Deze waterstand < 3.0 is, zit de waarde -9.0 in de database, om aan te geven dat de kering open blijft.

        # Interpoleer sluitpeilen op het grid van de windsnelheid en zeewaterstand
        # Eerst over de windsnelheid
        intstr = InterpStruct(
            x=wind_speed.get_discretisation(), xp=self.closing_levels.u
        )
        levels_u = intstr.interp(
            fp=self.closing_levels.h_rpb,
            axis=self.closing_levels.input_variables.index("u"),
        )

        # Dan over de zeewaterstand
        intstr = InterpStruct(
            x=sea_level.get_discretisation(), xp=self.closing_levels.m
        )
        self.h_rpb = intstr.interp(
            fp=levels_u, axis=self.closing_levels.input_variables.index("m")
        )

        # Controleer op waarden van -9 en kleiner.
        self.not_closing = self.h_rpb <= -9.0

    def calculate_closing_probability(
        self, wind_direction: float, closing_situation: int
    ) -> np.ndarray:
        """
        Bereken de kans op het al dan niet falen van N sluizen
        in de Oosterscheldekering en het type sluiting (nood of strategie)
        De kansen zijn conditioneel op de zeewaterstand.
        """
        # Translate to wind direction id
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)

        # Vertaal sluitsituatie (id) naar nood/strategie...
        sluittype = self.closing_situations[closing_situation][0]
        if sluittype not in ["Reguliere sluiting", "Noodsluiting"]:
            raise KeyError(sluittype)

        # ...en aantal schuiven
        nfaal = self.closing_situations[closing_situation][1]
        ik = self.k.tolist().index(nfaal)

        # Hernoem voor beter begrip
        kans_strategie_m = self.p_closing_k[:, 0]
        kans_handmatig_m = self.p_closing_k[:, 1]
        kans_noodsluit_m = self.p_closing_k[:, 2]
        # kans_geensluit_m = self.p_closing_k[:, 3]

        kansfalen_bemand = self.p_fail_k[:, 0]
        kansfalen_onbemand = self.p_fail_k[:, 1]

        # Kansen op noodsluiting, en N falende schuiven
        # Noodsluiting door mens (bemand) of systeem (onbemand)
        if sluittype == "Noodsluiting":
            kansfalensluiting = (
                kansfalen_bemand[ik] * kans_handmatig_m
                + kansfalen_onbemand[ik] * kans_noodsluit_m
            )

        elif sluittype == "Reguliere sluiting":
            # Strategiesluiting (bemand)
            kansfalensluiting = kansfalen_bemand[ik] * kans_strategie_m

            # En dan is er nog de kans dat de kering niet hoeft te sluiten (ongeacht bemand of onbemand)
            # Deze kans moet ingedeeld worden in een van de sluitscenario's. Neem hiervoor het sluitscenario
            # met zoveel mogelijk falende sluitingen, en de strategiesluiting
            # if nfaal == self.k.max():
            # kansfalensluiting[:] += kans_geensluit_m

        # Bereken de faalkansen aan de hand van de zeewaterstanden die voor de belastingcombinaties
        # optreden bij Roompot Buiten door te interpoleren op de zeewaterstanden
        keringsituatiekansen = Interpolate.inextrp1d(
            x=self.h_rpb[:, :, ir, :, :], xp=self.m_closing, fp=kansfalensluiting
        )

        # Als laatste is de kans op de keringsituatie nog afhankelijk van het al dan niet halen van het sluitpeil
        # Als het sluitpeil niet gehaald wordt, is de kans op de keringsituatie met sluiting 0, tenzij het om de
        # situatie zonder sluiting gaat
        idx = self.not_closing[:, :, ir, :, :]
        if (sluittype == "Reguliere sluiting") and (nfaal == self.k.max()):
            keringsituatiekansen[idx] = 1.0
        else:
            keringsituatiekansen[idx] = 0.0

        return keringsituatiekansen



import numpy as np

from ..statistics import Statistics
from ..stochastics.barrier.barrier_europoort import BarrierEuropoort
from ..stochastics.discharge import Discharge
from ..stochastics.discrete_probability import DiscreteProbability
from ..stochastics.model_uncertainty import ModelUncertainty
from ..stochastics.sea_level.sea_level_lower_river import SeaLevelLowerRiver
from ..stochastics.wind_speed import WindSpeed
from ....settings.settings import Settings
from .....common.interpolate import Interpolate
from .....common.probability import ProbabilityFunctions


class StatisticsLowerRiver(Statistics):
    """
    Statistics class for the Coast
    Water systems: Coast (North, Central, South), Waddensea (West, East) and Western scheldt
    """

    def __init__(self, settings: Settings):
        """
        Init the Statistics class

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit initialisation method from parent
        super().__init__(settings)

        # Sea level
        self.sea_level = SeaLevelLowerRiver(settings)

        # Wind
        self.wind_direction = DiscreteProbability(settings.wind_direction_probability)
        self.wind_speed = WindSpeed(settings)

        # Discharge
        self.discharge = Discharge(settings)

        # Europoort Barrier
        self.barrier = BarrierEuropoort(
            settings,
            self.wind_direction,
            self.wind_speed,
            self.sea_level,
            self.discharge,
        )

        # Model uncertainty
        self.model_uncertainties = ModelUncertainty(settings)

        # Discrete, slow, fast stochatics
        self.stochastics_discrete = {
            "r": self.wind_direction.get_discretisation(),
            "k": self.barrier.k,
        }
        self.stochastics_fast = {
            "u": self.wind_speed.get_discretisation(),
            "m": self.sea_level.get_discretisation(),
        }
        self.stochastics_slow = {"q": self.discharge.get_discretisation()}

    def calculate_probability(
        self, wind_direction: float, closing_situation: int = 1, given: list = []
    ):
        """
        Calculate the probability of occurence for the discretisation given the wind direction.

        Parameters
        ----------
        direction : float
            Wind direction
        closing_situation : int
            Closing situation, (irrelevant for Coast)
        given : list
            Given stochasts
        """
        # Sector West: Discharge, sea level and Europoort barrier
        if ((wind_direction >= 0.0) and (wind_direction <= 11.25)) or (
            (wind_direction > 212.75) and (wind_direction <= 360.0)
        ):
            if "q" in given:
                kansafvoer = np.ones_like(self.discharge.get_discretisation())
            else:
                # Note: use the momentary probability of the discharge
                kansafvoer = ProbabilityFunctions.probability_density(
                    self.discharge.get_discretisation(),
                    1 - self.discharge.get_exceedance_probability(),
                ).probability

            # Kans op windrichting
            ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
            kanswr = 1.0 if "r" in given else self.wind_direction.get_probability()[ir]

            # Bepaal sluitkansen voor de windrichting
            sluitkans = self.barrier.calculate_closing_probability(
                wind_direction, closing_situation
            )
            if "k" in given:
                sluitkans[:] = 1.0

            # Overschrijdingskans windsnelheid
            ovkansu_m = self.ovkansu_m(wind_direction)
            kanswind = np.array(
                [
                    ProbabilityFunctions.probability_density(
                        self.wind_speed.get_discretisation(), ovkansu_m[:, im]
                    ).probability
                    for im in range(len(self.sea_level))
                ]
            ).T[:, :, None]
            if "u" in given:
                kanswind[:] = 1.0

            # Kans zeewaterstand
            kanszws = ProbabilityFunctions.probability_density(
                self.sea_level.get_discretisation(),
                self.sea_level.get_exceedance_probability()[:, ir],
            ).probability[None, :, None]
            if "m" in given:
                kanszws[:] = 1.0

            # Combineer alle kansen
            comb = kanswind * kanszws * kansafvoer[None, None, :] * sluitkans * kanswr
            return np.swapaxes(comb, 1, 2)

        # Otherwise, sector east (without sea level and Europoort)
        else:
            # Kans windsnelheid
            ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
            kanswind = ProbabilityFunctions.probability_density(
                self.wind_speed.get_discretisation(),
                self.wind_speed.get_exceedance_probability()[:, ir],
            ).probability[:, None]
            if "u" in given:
                kanswind[:] = 1.0

            # Afvoer
            if "q" in given:
                kansafvoer = np.ones_like(self.discharge.get_discretisation())
            else:
                # Let op! gebruikt de MOMENTANE kans van een afvoer
                kansafvoer = ProbabilityFunctions.probability_density(
                    self.discharge.get_discretisation(),
                    1 - self.discharge.get_exceedance_probability(),
                ).probability

            # Kans op windrichting
            kanswr = 1.0 if "r" in given else self.wind_direction.get_probability()[ir]

            # Combineer alle kansen
            return kanswind * kansafvoer[None, :] * kanswr

    def ovkansu_m(self, wind_direction: float):
        """
        Bereken de overschrijdingskans van de windsnelheid volgens het
        correlatiemodel met de

        Parameters
        ----------
        richting : float
            Windrichting

        Returns
        -------
        ondkans : np.ndarray
            2D-array per windsnelheid en zeewaterstand
        """
        #  Bepaal transformatiewaarde van de windsnelheid voor het juiste rooster van windsnelheden
        if self.settings.transitional_wind == 0:
            kru = self.bereken_kru_polynoom(wind_direction)
        elif self.settings.transitional_wind == 1:
            kru = self.bereken_kru_tabel(wind_direction)
        else:
            raise ValueError(self.settings.transitional_wind)

        #  Bereken de windsnelheidkansen per richting en zeewaterstand voor de windsnelheid in de statistiek
        return 1.0 - self.ondkanswindsnelheid(
            wind_direction, kru, self.sea_level.get_discretisation()
        )

    def ondkanswindsnelheid(self, wind_direction, kru, m):
        """
        Bepaling van de onderschrijdingskans van de windsnelheid gegeven de
        zeewaterstand en de windrichting

        Parameters
        ----------
        wind_direction : float
            Windrichting
        kru : np.ndarray
            Transformatiewaarde van de windsnelheid

        Returns
        -------
        kans : np.array [len(kru), len(m)]
            Onderschrijdingskans van de windsnelheid gegeven de zeewaterstand en de windrichting
        """
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
        windparams = self.sea_level.pwinds[:, ir]

        # Bepaal de onderschrijdingskans van de windsnelheid uit de
        # Gumbelverdeelde getransformeerde waarde van de windsnelheid
        alpha = (
            windparams[5]
            * (m + self.sea_level.translation_m - windparams[0])
            / windparams[1]
        )
        ondkans = np.exp(-np.exp((-kru[:, None] + alpha[None, :]) / windparams[6]))

        # Afknotten van de Gumbelverdeling
        ondkans = np.minimum(1.0, (1.0 / (1.0 - self.settings.fu)) * ondkans)

        return ondkans

    def bereken_kru_tabel(self, wind_direction: float):
        """
        Berekenen van de transformatietabel van de windsnelheid

                Creeer eerst een vector met waarden voor het hele bereik van K_r(u)
        Ken vectorafmetingen toe aan parameters
        Alloceer geheugen voor het hele bereik van K_r(u)
        Maak lokaal een vector met zeewaterstanden
        Alloceer geheugen voor de zeewaterstanden
        Voor alle windsnelheden
        {
        Voor alle zeewaterstanden
        {
            Bepaal de onderschrijdingskans van K_r(u) gegeven de zeewaterstand
                en de windrichting
        }
        Bereken de integraal over de zeewaterstanden
        Verwijder de K_r(u)'s met onderschijdingskansen, die bij lagere waardes
            ook al voorkomen
        Voor kleine onder- en overschrijdingskansen de K_r(u) berekenen met
            extrapolatie
        }
        Geef het gealloceerde geheugen weer vrij
        """
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)

        #  Creeer eerst een vector met waarden voor het hele bereik van K_r(u)
        kru = np.arange(-3.0, 20.001, 0.2)

        #  Maak lokaal een vector met zeewaterstanden
        m_max = max(7.0, self.sea_level.get_discretisation()[-1])
        m = np.linspace(0.75, m_max, int(round((m_max - 0.75) / 0.05)) + 1)

        # Bepaal de onderschrijdingskans van K_r(u) gegeven de zeewaterstand en de windrichting

        # Bepaal de overschrijdingskansen van de zeewaterstand gegeven de windrichting op het
        # juiste rooster (N.B. kansdichtheden worden niet geenterpoleerd##)
        # TODO: logaritmisch?
        ovkanszws = Interpolate.inextrp1d(
            x=m,
            xp=self.sea_level.get_discretisation(),
            fp=self.sea_level.get_exceedance_probability()[:, ir],
        )
        ovkanszws[ovkanszws > 1.0] = 1.0
        ovkanszws = np.maximum.accumulate(ovkanszws[::-1])[::-1]

        # Bereken de kansdichtheid uit de overschrijdingskansen
        kanszws = ProbabilityFunctions.probability_density(m, ovkanszws).probability

        # Bereken de integraal over de zeewaterstanden
        ondkans_y_r = (
            self.ondkanswindsnelheid(wind_direction, kru, m) * kanszws[None, :]
        ).sum(1)

        # Verwijder de K_r(u)'s met onderschijdingskansen, die bij lagere waardes ook al voorkomen
        ondkans_uniek, idx = np.unique(ondkans_y_r, return_index=True)
        kru_uniek = kru[idx]

        # Voor kleine onder- en overschrijdingskansen de K_r(u) berekenen met extrapolatie
        # Bepaal eerste het gebied dat BINNEN deze grenzen valt
        ovkans_u = self.wind_speed.get_exceedance_probability()[:, ir]
        idx = (ovkans_u >= 1.0e-7) & (ovkans_u <= 1 - 1.0e-8)

        # Alloceer nieuwe array voor kru
        kru = np.zeros_like(ovkans_u)

        # Interpoleer dit gebied
        kru[idx] = Interpolate.inextrp1d(
            x=1.0 - ovkans_u[idx], xp=ondkans_uniek, fp=kru_uniek
        )

        # Extrapoleer de andere waarden hier omheen
        kru[~idx] = Interpolate.inextrp1d(
            x=self.wind_speed.get_discretisation()[~idx],
            xp=self.wind_speed.get_discretisation()[idx],
            fp=kru[idx],
        )

        return kru

    def bereken_kru_polynoom(self, wind_direction: float):
        """
        Bepaal transformatiewaarden van de windsnelheden

        Parameters
        ---------

        Returns
        -------
        kru
            Vector met transformatiewaarden voor de windsnelheid op het gewenste rooster van de windsnelheden
        """
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
        windparams = self.sea_level.pwinds[:, ir]

        #  Bepaal transformatiewaarden van de windsnelheden
        kru = (
            windparams[2] * self.wind_speed.get_discretisation() ** 2
            + windparams[3] * self.wind_speed.get_discretisation()
            + windparams[4]
        )

        return kru



import numpy as np

from .sea_level import SeaLevel
from .....settings.settings import Settings
from ......common.interpolate import Interpolate
from ......common.probability import ProbabilityFunctions
from ......io.file_hydranl import FileHydraNL


class SeaLevelPoint(SeaLevel):
    """
    Class to describe the sea (water) level statistics.
    Sea level statistics are conditional on the wind direction.
    """

    def __init__(self, settings: Settings, epsilon: float = 1.0e-15):
        """
        For Eastern Scheldt systems, we base the SeaLevel statistics on one
        point.

        Parameters
        ----------
        settings : Settings
            The Settings object
        epsilon : float
            Maximum exceedance probability to determine the transformation
            table, used to take correlation into account.
        """
        # Inherit
        super().__init__()

        #  Lees de conditionele overschrijdingskansen van de zeewaterstand gegeven de windrichting
        m, epm = FileHydraNL.read_file_ncolumns(settings.sea_level_probability)
        nr = epm.shape[1]

        # Equidistant vullen van vector met zeewaterstanden
        settings.m_min = settings.m_min if settings.m_min is not None else m[0]
        self.m = ProbabilityFunctions.get_hnl_disc_array(
            settings.m_min, settings.m_max, settings.m_step
        )

        # Alloceer matrices
        self.nm = len(self.m)

        #  Interpoleer de conditionele overschrijdingskansen van de zeewaterstand gegeven de windrichting naar het gewenste rooster
        self.epm = np.zeros((self.nm, nr))
        for ir in range(nr):
            self.epm[:, ir] = np.maximum(
                0, np.exp(Interpolate.inextrp1d(self.m, m, np.log(epm[:, ir])))
            )

        # Adjust for sea level rise
        self.m = self.m + settings.sea_level_rise

        #  Berekenen kansdichtheid zeewaterstand gegeven de windrichting
        self.pm = np.zeros((self.nm, nr))
        for ir in range(nr):
            self.pm[:, ir] = ProbabilityFunctions.probability_density(
                self.m, self.epm[:, ir]
            ).density

        # Bereken de transformatietabel van de OVERschrijdingskansen naar exponentiële ruimte
        # voor het correlatiemodel van de zeewaterstand en de windsnelheid.
        # Zie ook vgl. 4.7 van [Geerse, 2012], merk op dat het daar ONDERschrijdingskansen betreft#
        self.epm_exp = -np.log(np.maximum(self.epm, epsilon))



import os
import pandas as pd
import sqlite3

from ..common.enum import WaterSystem


class DatabaseSettings:
    """
    Settings database
    """

    def __init__(self) -> None:
        self.con = None

    def __enter__(self) -> "DatabaseSettings":
        # Init the connection
        self.con = sqlite3.connect(
            os.path.join(
                os.path.split(os.path.dirname(__file__))[0],
                "data",
                "settings",
                "calculation_settings.sqlite",
            )
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # Close the connection
        self.con.close()

    def get_settings(self, watersystem: WaterSystem) -> pd.DataFrame:
        """
        Return all settings from the database

        Parameters
        ----------
        watersystem : WaterSystem
            WaterSystem to get the settings for

        Returns
        -------
        pd.DataFrame
            A DataFrame with settings
        """
        # Query
        sql = f"SELECT SettingName, SettingValue FROM CalculationSettings WHERE WaterSystem = {watersystem.value}"
        settings = pd.read_sql(sql, self.con)

        # Parse to int and float if possible
        for n, row in settings.iterrows():
            settings.loc[n, "SettingValue"] = self.__parse_value(row["SettingValue"])

        # Return the settings
        return settings

    def get_sea_level_statistic_points(self, watersystem: WaterSystem) -> pd.DataFrame:
        """
        Return the reference points for sea level statistics

        Parameters
        ----------
        watersystem : WaterSystem
            Watersystem for which to return the reference points

        Returns
        -------
        pd.DataFrame
            All reference points
        """
        # Obtain the right table name
        if watersystem in [
            WaterSystem.COAST_SOUTH,
            WaterSystem.COAST_CENTRAL,
            WaterSystem.COAST_NORTH,
        ]:
            tablename = "CoastReferencePoints"
        elif watersystem in [WaterSystem.WADDEN_SEA_WEST, WaterSystem.WADDEN_SEA_EAST]:
            tablename = "WaddenSeaReferencePoints"
        else:
            raise NotImplementedError(
                f"[ERROR] No sea level reference points implemented for {watersystem}."
            )

        # SQL
        data = pd.read_sql(
            f"SELECT Name, X, Y FROM {tablename}", self.con, index_col="Name"
        )

        # Return data
        return data

    def get_sea_level_sub_systems(self, watersystem: WaterSystem) -> pd.DataFrame:
        """
        Return all subsystems of the coast or Wadden sea.

        Parameters
        ----------
        watersystem : WaterSystem
            Watersystem for which to return the reference points

        Returns
        -------
        pd.DataFrame
            All subsystems
        """
        # Obtain the right table name
        if watersystem in [
            WaterSystem.COAST_SOUTH,
            WaterSystem.COAST_CENTRAL,
            WaterSystem.COAST_NORTH,
        ]:
            tablename = "CoastSubSystems"
        elif watersystem in [WaterSystem.WADDEN_SEA_WEST, WaterSystem.WADDEN_SEA_EAST]:
            tablename = "WaddenSeaSubSystems"
        else:
            raise NotImplementedError(
                f"[ERROR] No sea level reference points implemented for {watersystem}."
            )

        # SQL
        data = pd.read_sql(
            f"SELECT SubSystemId, Point1, Point2, Point3 FROM {tablename}",
            self.con,
            index_col="SubSystemId",
        )

        # Return data
        return data

    def get_input_variable_ids(self) -> dict:
        """
        Return a dictionary with the input variable ids (ivid) and symbols (isymbol)

        Returns
        -------
        dict
            A dictionary with {ivid : isymbol}
        """
        # Query
        sql = "SELECT InputVariableId, InputVariableSymbol FROM InputVariables"
        data = self.con.execute(sql).fetchall()

        # Process data
        results = {data[i][0]: data[i][1] for i in range(len(data))}

        # Return the dictionary
        return results

    def get_result_variable_ids(self) -> dict:
        """
        Return a dictionary with the result variable ids (rvid) and symbols (rsymbol)

        Returns
        -------
        dict
            A dictionary with {rvid : rsymbol}
        """
        # Query
        sql = "SELECT ResultVariableId, ResultVariableSymbol FROM ResultVariables"
        data = self.con.execute(sql).fetchall()

        # Process data
        results = {data[i][0]: data[i][1] for i in range(len(data))}

        # Return the dictionary
        return results

    def __parse_value(self, value):
        # Parse value from string
        try:
            return float(value)
        except ValueError:
            pass
        try:
            return int(value)
        except ValueError:
            pass
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        return value



import numpy as np

from dataclasses import dataclass
from pathlib import Path

from ...common.interpolate import Interpolate


@dataclass
class FrequencyLine:
    """
    Represents a frequency line with associated levels and exceedance frequencies.

    Parameters
    ----------
    level : np.ndarray
        An array of levels.
    exceedance_frequency : np.ndarray
        An array of corresponding exceedance frequencies.
    """

    # Init variables
    level: np.ndarray
    exceedance_frequency: np.ndarray

    def interpolate_exceedance_probability(self, exceedance_probability: np.ndarray):
        exceedance_probability = np.atleast_1d(exceedance_probability)
        order_x = np.argsort(exceedance_probability)
        order_xp = np.argsort(self.exceedance_frequency)
        f = Interpolate.inextrp1d(
            x=np.log(exceedance_probability)[order_x],
            xp=np.log(self.exceedance_frequency)[order_xp],
            fp=self.level[order_xp],
        )
        return f[np.argsort(order_x)]

    def interpolate_level(self, level: np.ndarray):
        return np.exp(
            Interpolate.inextrp1d(
                x=level, xp=self.level, fp=np.log(self.exceedance_frequency)
            )
        )

    def to_file(self, path: Path, overwrite=False):
        if not overwrite and path.exists():
            raise OSError(
                f'Path "{path}" already exists. Choose overwrite=True or give another path.'
            )

        with path.open("w") as f:
            f.write(f"{len(self.level):5d}\n")
            for niveau, ovfreq in zip(self.level, self.exceedance_frequency):
                f.write(f"{niveau:11.6f} {ovfreq:14.7e}\n")

    def drop_zeros(self):
        idx = self.exceedance_frequency == 0.0
        if idx.all():
            raise ValueError("All exceedance frequencies are 0.")
        self.exceedance_frequency = self.exceedance_frequency[~idx]
        self.level = self.level[~idx]



import numpy as np

from scipy.stats import norm

from .barrier import Barrier
from ..discharge import Discharge
from ..discrete_probability import DiscreteProbability
from ..sea_level.sea_level import SeaLevel
from ..wind_speed import WindSpeed
from ....loading.loading_model.loading_model import LoadingModel
from .....settings.settings import Settings
from ......common.interpolate import InterpStruct
from ......io.database_hr import DatabaseHR


class BarrierEuropoort(Barrier):
    def __init__(
        self,
        settings: Settings,
        wind_direction: DiscreteProbability,
        wind_speed: WindSpeed,
        sea_level: SeaLevel,
        discharge: Discharge,
    ):
        # Inherit
        super().__init__(settings, wind_direction)

        # Save statistics
        self.discharge = discharge
        self.sea_level = sea_level
        self.wind_speed = wind_speed

        # Assert barrier == 1
        if self.settings.europoort_barrier != 1:
            raise NotImplementedError("[ERROR] Barrier != 1 not implemented.")

        # Init standard variables
        self.k = [0, 11]
        self.nk = len(self.k)

        # Database connection
        with DatabaseHR(settings.database_path) as database:
            # Closing levels
            self.closing_levels = LoadingModel(None, None, ["u", "q", "r"], ["m"])
            self.closing_levels.initialise(
                database.get_closing_levels_table_europoort()
            )

        # Extend
        self.closing_levels.extend("q", [self.settings.q_min, self.settings.q_max])
        self.closing_levels.extend("u", [self.settings.u_max])

        # Prepare probabilities
        self.prepare_barrier_probability(sea_level)

    def calculate_closing_probability(
        self, wind_direction: float, closing_situation: int = None
    ) -> np.ndarray:
        # Translate to wind direction id
        ir = self.wind_direction.get_discretisation().tolist().index(wind_direction)
        nqblok = len(self.discharge)
        nmblok = len(self.sea_level)

        # Bepaal de index van de richting in de sluitpeilen
        if wind_direction not in self.closing_levels.r:
            raise KeyError(
                f"[ERROR] Wind direction {wind_direction} not defined in closing levels loading."
            )

        ir = self.closing_levels.r.tolist().index(wind_direction)
        if closing_situation == 1:  # Open
            ik = 0  # 00
            sluitkans = np.zeros((len(self.wind_speed), nmblok, nqblok))
        elif closing_situation == 2:  # Closed
            ik = 1  # 11
            sluitkans = np.zeros((len(self.wind_speed), nmblok, nqblok))
        elif closing_situation is None:  # Both
            ik = slice(None)
            sluitkans = np.zeros((len(self.wind_speed), nmblok, nqblok, len(self.k)))
        else:
            raise KeyError(closing_situation)

        # De sluitpeilen (een zeewaterstand) zijn bekend per windsnelheid, windrichting en afvoer
        # Eerste worden de sluitpeilen uitgebreid voor alle windsnelheden en afvoerniveaus in de statistiek

        # Interpoleer over de windsnelheden, de eerste as van peilen
        sluitpeil = InterpStruct(
            x=self.wind_speed.get_discretisation(), xp=self.closing_levels.u
        ).interp(fp=self.closing_levels.m[:, :, ir], axis=0)

        # Interpoleer dit resultaat over de afvoeren, de tweede as van de peilen
        sluitpeil = InterpStruct(
            x=self.discharge.get_discretisation(), xp=self.closing_levels.q
        ).interp(fp=sluitpeil, axis=1)

        # Daarna wordt de sluitkans bepaald, door de zeewaterstanden bij deze windsnelheid, zeewaterstand en keringsituatie
        # te interpoleren in de berekende sluitkansen per cm.

        # De sluitkansen zijn afhankelijk van het sluitpeil, de zeewaterstand en (uiteraard) de sluitsituatie,
        # de afmetingen van self.kansen is daarmee ncmsp, nmblok, nk

        # De sluitkans moet bepaald worden per windsnelheid, zeewaterstand, afvoer en keringsituatie

        # berekend voor de in de database aanwezig windsnelheden, windrichting, en elke centimeter zeewaterstand

        # Voor elke afvoer, interpoleer de kansen
        for iq in range(nqblok):
            # bepaal de sluitkansen voor de combinatie van afvoer, kering en zeewaterstand
            if isinstance(ik, int):
                sluitkans[:, :, iq] = InterpStruct(
                    x=sluitpeil[:, iq], xp=self.cmsp
                ).interp(self.kansen[:, :, ik], axis=0)
            else:
                sluitkans[:, :, iq, :] = InterpStruct(
                    x=sluitpeil[:, iq], xp=self.cmsp
                ).interp(self.kansen[:, :, :], axis=0)

        return sluitkans

    def prepare_barrier_probability(self, sea_level: SeaLevel):
        """
        bereken het minimale sluitpeil, op cm's naar beneden afgerond

        bepaal het aantal cm's van het minimale tot het maximale sluitpeil
        alloceer geheugen

        voor elke cm tussen het minimum en maximum
            bepaal de hoogte van het sluitpeil

        voor elke keringsituatie
            bereken de keringskans per zeewaterstand en sluitpeil

        Parameters
        ----------
        belasting : type (tprand2001)
            structure met alle sluitpeilen
        statistiek : type (tpstatistiek)
            structure met waterstanden, golfparameters en hydraulische belastingen als gevolg daarvan
        """
        # bereken het minimale sluitpeil, op cm's naar beneden afgerond
        step_size = 0.01
        min_cl = np.floor(self.closing_levels.m.min() / step_size) * step_size
        max_cl = np.ceil(self.closing_levels.m.max() / step_size) * step_size

        # bepaal het aantal cm's van het minimale tot en met het maximale sluitpeil
        self.nsp = int((max_cl - min_cl) / step_size) + 1

        # alloceer geheugen
        self.kansen = np.zeros((self.nsp, len(sea_level), len(self.k)))

        # vul de vector met sluitpeilen voor elke cm
        self.cmsp = min_cl + np.arange(self.nsp) * step_size

        # Settings
        barrier = int(self.settings.europoort_barrier)
        distribution = self.settings.europoort_barrier_distribution
        mu = self.settings.europoort_barrier_mu
        sigma = self.settings.europoort_barrier_sigma
        alfa = self.settings.europoort_barrier_alfa

        # bereken de keringskans voor elke keringtoestand, zeewaterstand en sluitpeilcentimeter
        for ik in range(self.nk):
            #  bepaal kans op commando keringen dicht
            if barrier != 2:
                #  Normal distribution
                if distribution == 0:
                    s = (
                        self.cmsp[:, None]
                        - sea_level.get_discretisation()[None, :]
                        - mu
                    ) / sigma
                    kansdicht = 1.0 - norm(loc=0.0, scale=1.0).cdf(x=s)

                # Cosinus-squared distribution
                elif distribution == 1:
                    s = np.pi * sigma * (3.0 / (np.pi * np.pi - 6.0)) ** 0.5
                    kansdicht = 0.5 * (
                        1.0
                        + (
                            self.cmsp[:, None]
                            - sea_level.get_discretisation()[None, :]
                            - mu
                        )
                        / s
                        + np.sin(
                            np.pi
                            * (
                                self.cmsp[:, None]
                                - sea_level.get_discretisation()[None, :]
                                - mu
                            )
                            / s
                        )
                        / np.pi
                    )

                # Undefined distribution
                else:
                    raise ValueError("Fout type kansverdeling voorspelde zeewaterstand")

            if barrier == 1:
                if int(round(self.k[ik])) == 0:
                    self.kansen[:, :, ik] = 1.0 - (1.0 - alfa) * kansdicht
                elif int(round(self.k[ik])) == 11:
                    self.kansen[:, :, ik] = (1.0 - alfa) * kansdicht

            # Assert barrier == 1 (for now)
            else:
                raise NotImplementedError("[ERROR] Barrier != 1 not implemented.")



import ast
import matplotlib.pyplot as plt
import numpy as np
import os

from typing import Tuple, Union

from .foreland import Foreland
from .profile_loading import ProfileLoading
from ...common.enum import Breakwater
from ...common.interpolate import Interpolate


class Profile:
    """
    Profile class

    TODO: Implement support for sheet piles
    """

    # Profile name
    profile_name = None

    # Breakwater
    breakwater_type = Breakwater.NO_BREAKWATER
    breakwater_level = 0.0

    # Foreland
    foreland_x_coordinates: None | list = None
    foreland_y_coordinates: None | list = None

    # Dike schematisation
    dike_orientation: None | float = None
    dike_crest_level: None | float = None
    dike_x_coordinates: None | list = None
    dike_y_coordinates: None | list = None
    dike_roughness: None | list = None

    def __init__(self, profile_name: str = "Profile"):
        """
        Create a new profile

        Parameters
        ----------
        profile_name: str
            Name of the profile (default: 'Profile')
        """
        # Save the profile name
        self.profile_name = profile_name

    def validate_profile(self) -> bool:
        """
        Function to validate the profile.

        Returns
        -------
        bool
            True if the profile is OK
        """
        # TODO: Improve validation
        # Check necessary components
        # Orientation
        if self.dike_orientation is None:
            return False

        # Crest height
        if self.dike_crest_level is None:
            return False

        # Geometry
        if (
            (self.dike_x_coordinates is None)
            or (self.dike_y_coordinates is None)
            or (self.dike_roughness is None)
        ):
            return False

        return True

    def set_dike_orientation(self, dike_orientation: float):
        """
        Change the dike orientation

        Parameters
        ----------
        dike_orientation : float
            The dike orientation
        """
        self.dike_orientation = dike_orientation

    def set_dike_crest_level(self, dike_crest_level: float):
        """
        Change the crest level

        Parameters
        ----------
        dike_crest_level : float
            The crest level of the dike
        """
        self.dike_crest_level = dike_crest_level

    def set_dike_geometry(
        self,
        dike_x_coordinates: list,
        dike_y_coordinates: list,
        dike_roughness: list = None,
    ):
        """
        Change the geometry of the outer slope of the dike

        Parameters
        ----------
        dike_x_coordinates : list
            A list with the x coordinates of the profile
        dike_y_coordinates : list
            A list with the y coordinates of the profile
        dike_roughness : list, optional
            A list with the roughness of the profile (default : all parts 1.0)
        """
        # Make sure the dike geometry starts at (x = 0)
        min_x = np.min(dike_x_coordinates)

        # If dike roughness is None, the roughness of every part is 1.0
        if dike_roughness is None:
            dike_roughness = [1.0] * len(dike_x_coordinates)

        # Save the dike geometry coordinates
        self.dike_x_coordinates = dike_x_coordinates - min_x
        self.dike_y_coordinates = dike_y_coordinates
        self.dike_roughness = dike_roughness

    def has_foreland(self) -> bool:
        """
        Returns whether the profile has a foreland or breakwater

        Returns
        -------
        bool
            True if the profile has a foreland or breakwater
        """
        # Check for a breakwater
        if self.breakwater_type != Breakwater.NO_BREAKWATER:
            return True

        # Check for a foreland
        if self.foreland_x_coordinates is not None:
            return True

        # No foreland or breakwater
        return False

    def set_foreland_geometry(
        self, foreland_x_coordinates: list = None, foreland_y_coordinates: list = None
    ):
        """
        Change the geometry of the foreland
        Setting the foreland x and y coordinates to None will remove the foreland.

        Parameters
        ----------
        foreland_x_coordinates : list
            A list with the x coordinates of the foreland (default: None)
        foreland_y_coordinates : list
            A list with the y coordinates of the foreland (default: None)
        """
        # Add foreland
        if foreland_x_coordinates is not None:
            # Make sure the foreland ends at (x = 0)
            max_x = np.max(foreland_x_coordinates)

            # Save the foreland coordinates
            self.foreland_x_coordinates = foreland_x_coordinates - max_x
            self.foreland_y_coordinates = foreland_y_coordinates

        # Delete foreland
        else:
            self.foreland_x_coordinates = None
            self.foreland_y_coordinates = None

    def remove_foreland(self) -> None:
        """
        Remove the foreland
        Wrapper for set_foreland_geometry()
        """
        self.set_foreland_geometry(
            foreland_x_coordinates=None, foreland_y_coordinates=None
        )

    def set_breakwater(
        self, breakwater_type: Breakwater = None, breakwater_level: float = 0.0
    ):
        """
        Change the breakwater
        """
        # None equals no breakwater
        if breakwater_type is None:
            breakwater_type = Breakwater.NO_BREAKWATER

        # If there is no breakwater, the height will be set to None
        if breakwater_type == Breakwater.NO_BREAKWATER:
            breakwater_level = 0.0

        # Apply settings
        self.breakwater_type = breakwater_type
        self.breakwater_level = breakwater_level

    def remove_breakwater(self) -> None:
        """
        Remove the breakwater
        Wrapper for set_breakwater()
        """
        self.set_breakwater(
            breakwater_type=Breakwater.NO_BREAKWATER, breakwater_level=0.0
        )

    def calculate_overtopping(
        self,
        water_level: Union[float, list],
        significant_wave_height: Union[float, list],
        spectral_wave_period: Union[float, list],
        wave_direction: Union[float, list],
        tp_tspec: float = 1.1,
        dll_settings: dict = None,
    ) -> Union[float, list]:
        """
        Calculate the overtopping discharge

        Parameters
        ----------
        water_level : Union[float, list]
            List or float with the water level
        significant_wave_height : Union[float, list]
            List or float with the significant wave height
        spectral_wave_period : Union[float, list]
            List or float with the spectral wave period
        wave_direction : Union[float, list]
            List or float with the wave direction
        tp_tspec : float, optional
            Ratio between Tp and Tspec, only used for dam and foreland (default : 1.1)

        Returns
        -------
        Union[float, list]
            List or float with the overtopping discharge
        """
        # Transform wave conditions
        water_level, significant_wave_height, spectral_wave_period, wave_direction = (
            self.transform_wave_conditions(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
                tp_tspec,
                force_array=True,
            )
        )

        # Create profile loading
        profile_loading = ProfileLoading(self, dll_settings)

        # Calculate
        qov = []
        for _h, _hs, _tspec, _dir in zip(
            water_level, significant_wave_height, spectral_wave_period, wave_direction
        ):
            qov.append(profile_loading.calculate_discharge(_h, _hs, _tspec, _dir))

        # Return
        return qov[0] if len(qov) == 1 else qov

    def calculate_runup(
        self,
        water_level: Union[float, list],
        significant_wave_height: Union[float, list],
        spectral_wave_period: Union[float, list],
        wave_direction: Union[float, list],
        tp_tspec: float = 1.1,
        dll_settings: dict = None,
    ) -> Union[float, list]:
        """
        Calculate the runup height

        Parameters
        ----------
        water_level : Union[float, list]
            List or float with the water level
        significant_wave_height : Union[float, list]
            List or float with the significant wave height
        spectral_wave_period : Union[float, list]
            List or float with the spectral wave period
        wave_direction : Union[float, list]
            List or float with the wave direction
        tp_tspec : float, optional
            Ratio between Tp and Tspec, only used for dam and foreland (default : 1.1)

        Returns
        -------
        Union[float, list]
            List or float with the runup height
        """
        # Transform wave conditions
        water_level, significant_wave_height, spectral_wave_period, wave_direction = (
            self.transform_wave_conditions(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
                tp_tspec,
                force_array=True,
            )
        )

        # Create profile loading
        profile_loading = ProfileLoading(self, dll_settings)

        # Calculate
        ru2p = []
        for _h, _hs, _tspec, _dir in zip(
            water_level, significant_wave_height, spectral_wave_period, wave_direction
        ):
            ru2p.append(profile_loading.calculate_runup(_h, _hs, _tspec, _dir))

        # Return
        return ru2p[0] if len(ru2p) == 1 else ru2p

    def calculate_crest_level(
        self,
        q_overtopping: float,
        water_level: Union[float, list],
        significant_wave_height: Union[float, list],
        spectral_wave_period: Union[float, list],
        wave_direction: Union[float, list],
        tp_tspec: float = 1.1,
        dll_settings: dict = None,
    ) -> Union[float, list]:
        """
        Calculate the crest level for a given overtopping discharge

        Parameters
        ----------
        q_overtopping : float
            Critical overtopping discharge
        water_level : Union[float, list]
            List or float with the water level
        significant_wave_height : Union[float, list]
            List or float with the significant wave height
        spectral_wave_period : Union[float, list]
            List or float with the spectral wave period
        wave_direction : Union[float, list]
            List or float with the wave direction
        tp_tspec : float, optional
            Ratio between Tp and Tspec, only used for dam and foreland (default : 1.1)

        Returns
        -------
        Union[float, list]
            List or float with the crest level
        """
        # Transform wave conditions
        water_level, significant_wave_height, spectral_wave_period, wave_direction = (
            self.transform_wave_conditions(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
                tp_tspec,
                force_array=True,
            )
        )

        # Length of the list
        n_q = 1 if isinstance(q_overtopping, (float, int)) else len(q_overtopping)

        # If q_overtopping is an int or float
        if isinstance(q_overtopping, (float, int)):
            q_overtopping = [q_overtopping] * len(water_level)

        # If q_overtopping is a list and water_level has length 1
        elif n_q > 1 and len(water_level) == 1:
            water_level = [water_level] * n_q
            significant_wave_height = [significant_wave_height] * n_q
            spectral_wave_period = [spectral_wave_period] * n_q
            wave_direction = [wave_direction] * n_q

        # Uneven length of arrays
        else:
            raise ValueError("[ERROR] Uneven length of arrays")

        # To numpy
        q_overtopping = np.array(q_overtopping)
        water_level = np.array(water_level)
        significant_wave_height = np.array(significant_wave_height)
        spectral_wave_period = np.array(spectral_wave_period)
        wave_direction = np.array(wave_direction)

        # Create profile loading
        profile_loading = ProfileLoading(self, dll_settings)

        # Calculate crest level
        hbn = []
        for _q, _h, _hs, _tspec, _dir in zip(
            q_overtopping,
            water_level,
            significant_wave_height,
            spectral_wave_period,
            wave_direction,
        ):
            hbn.append(profile_loading.calculate_crest_level(_q, _h, _hs, _tspec, _dir))

        # Return
        return hbn[0] if len(hbn) == 1 else hbn

    def transform_wave_conditions(
        self,
        water_level: Union[float, list],
        significant_wave_height: Union[float, list],
        spectral_wave_period: Union[float, list],
        wave_direction: Union[float, list],
        tp_tspec: float = 1.1,
        force_array: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Transform the wave conditions for the schematized foreland

        Parameters
        ----------
        water_level : Union[float, list]
            Water level
        significant_wave_height : Union[float, list]
            Significant wave height
        spectral_wave_period : Union[float, list]
            Spectral wave period
        wave_direction : Union[float, list]
            Wave direction
        tp_tspec : float, optional
            Ratio between Tp and Tspec, only used for dam and foreland (default : 1.1)
        force_array : bool, optional
            Always force to return an array

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            Water level and transformed wave conditions (h, hs, tp, dir)
        """
        # Length of the lists
        n_h = 1 if isinstance(water_level, (float, int)) else len(water_level)
        n_hs = (
            1
            if isinstance(significant_wave_height, (float, int))
            else len(significant_wave_height)
        )
        n_tspec = (
            1
            if isinstance(spectral_wave_period, (float, int))
            else len(spectral_wave_period)
        )
        n_dir = 1 if isinstance(wave_direction, (float, int)) else len(wave_direction)
        n_max = np.max([n_h, n_hs, n_tspec, n_dir])

        # Make lists
        if isinstance(water_level, (float, int)):
            water_level = [water_level] * n_max
        elif len(water_level) != n_max:
            raise ValueError("[ERROR] Uneven length of arrays")

        if isinstance(significant_wave_height, (float, int)):
            significant_wave_height = [significant_wave_height] * n_max
        elif len(significant_wave_height) != n_max:
            raise ValueError("[ERROR] Uneven length of arrays")

        if isinstance(spectral_wave_period, (float, int)):
            spectral_wave_period = [spectral_wave_period] * n_max
        elif len(spectral_wave_period) != n_max:
            raise ValueError("[ERROR] Uneven length of arrays")

        if isinstance(wave_direction, (float, int)):
            wave_direction = [wave_direction] * n_max
        elif len(wave_direction) != n_max:
            raise ValueError("[ERROR] Uneven length of arrays")

        # To numpy
        water_level = np.array(water_level)
        significant_wave_height = np.array(significant_wave_height)
        spectral_wave_period = np.array(spectral_wave_period)
        wave_direction = np.array(wave_direction)

        # Correct for foreland
        if self.has_foreland():
            fl = Foreland(self)
            water_level, significant_wave_height, peak_wave_period, wave_direction = (
                fl.transform_wave_conditions(
                    water_level,
                    significant_wave_height,
                    spectral_wave_period * tp_tspec,
                    wave_direction,
                )
            )
            spectral_wave_period = peak_wave_period / tp_tspec

        # Return array or floats?
        if len(water_level.ravel()) == 1 and not force_array:
            return (
                water_level.ravel()[0],
                significant_wave_height.ravel()[0],
                spectral_wave_period.ravel()[0],
                wave_direction.ravel()[0],
            )
        else:
            return (
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
            )

    def to_prfl(
        self, export_path: str, id: str = "Onbekend000", memo: str = ""
    ) -> None:
        """
        Export to a prfl file

        export_path : str
            Path to where the profile has to be exported
        id : str
            Id used in RisKeer
        memo : str
            Memo to be added to the prfl file
        """
        # Check required info
        if (
            (self.dike_x_coordinates is None)
            or (self.dike_crest_level is None)
            or (self.dike_orientation is None)
        ):
            raise ValueError(
                f"[ERROR] Cannot generate .prfl for profile '{self.profile_name}', geometry, crest level or orientation is missing."
            )

        # Version
        breakwater_level = (
            self.breakwater_level if self.breakwater_level is not None else 0.0
        )
        n_foreland = (
            len(self.foreland_x_coordinates)
            if self.foreland_x_coordinates is not None
            else 0
        )
        n_dike = (
            len(self.dike_x_coordinates) if self.dike_x_coordinates is not None else 0
        )
        export = f"VERSIE 4.0\nID {id}\n\nRICHTING {self.dike_orientation}\n\nDAM {int(self.breakwater_type.value)}\nDAMHOOGTE {breakwater_level}\n\nVOORLAND {n_foreland}\n[FORELAND]\nDAMWAND 0\nKRUINHOOGTE {self.dike_crest_level}\nDIJK {n_dike}\n[DIKE]\nMEMO\nGenerated with Pydra for profile '{self.profile_name}'\n{memo}\n"

        # Foreland
        foreland = ""
        if n_foreland > 0:
            for foreland_x, foreland_y in zip(
                self.foreland_x_coordinates, self.foreland_y_coordinates
            ):
                foreland = (
                    foreland
                    + f"{round(foreland_x,3):.3f}\t{round(foreland_y,3):.3f}\t1.000\n"
                )
        export = export.replace("[FORELAND]", foreland)

        # Dike
        dike = ""
        if n_dike > 0:
            for dike_x, dike_y, dike_r in zip(
                self.dike_x_coordinates, self.dike_y_coordinates, self.dike_roughness
            ):
                dike = (
                    dike
                    + f"{round(dike_x,3):.3f}\t{round(dike_y,3):.3f}\t{round(dike_r,3):.3f}\n"
                )
        export = export.replace("[DIKE]", dike)

        # Export .prfl
        if not export_path.lower().endswith(".prfl"):
            export_path = export_path + ".prfl"
        prfl_file = open(export_path, "w")
        prfl_file.write(export)
        prfl_file.close()

    def to_dict(self) -> dict:
        """
        Export Profile to dictionary

        Returns
        -------
        dictionary
            Dictionary with all profile settings
        """
        # Create an empty dictionary
        profile = {}

        # Add all settings to the dictionary
        for setting in dir(self):
            if not str(setting).startswith("__") and not callable(
                getattr(self, setting)
            ):
                profile[setting] = getattr(self, setting)

        return profile

    def draw_profile(self) -> None:
        """
        Draw a cross-section of the profile
        """
        # Init plot
        plt.figure(figsize=[8, 5])

        # Foreland
        if self.foreland_x_coordinates is not None:
            plt.plot(
                self.foreland_x_coordinates,
                self.foreland_y_coordinates,
                color="orange",
                label="Foreland",
                zorder=1,
            )

        # Geometry
        for i in range(len(self.dike_x_coordinates) - 1):
            r = self.dike_roughness[i]
            plt.plot(
                self.dike_x_coordinates[i : i + 2],
                self.dike_y_coordinates[i : i + 2],
                color="black" if r != 1.0 else "green",
                label=f"Roughness: {r}",
                zorder=1,
            )

        # Breakwater
        if isinstance(self.breakwater_type, Breakwater):
            # Where should we draw the breakwater?
            if self.foreland_x_coordinates is not None:
                x_orig, y_orig = (
                    self.foreland_x_coordinates[0],
                    self.foreland_y_coordinates[0],
                )
            else:
                x_orig, y_orig = self.dike_x_coordinates[0], self.dike_y_coordinates[0]

            # Draw
            bw_height = self.breakwater_level
            if self.breakwater_type == Breakwater.CAISSON:
                plt.fill_between(
                    [x_orig - 10, x_orig],
                    [bw_height, bw_height],
                    [y_orig, y_orig],
                    color="dimgrey",
                    label="Caisson",
                    zorder=2,
                )
            elif self.breakwater_type == Breakwater.VERTICAL_WALL:
                plt.plot(
                    [x_orig, x_orig],
                    [y_orig, bw_height],
                    color="black",
                    lw=3,
                    label="Vertical wall",
                    zorder=2,
                )
            elif self.breakwater_type == Breakwater.RUBBLE_MOUND:
                diff = bw_height - y_orig
                plt.fill_between(
                    [
                        x_orig - (3 * diff) - 2,
                        x_orig - (1.5 * diff) - 2,
                        x_orig - (1.5 * diff),
                        x_orig,
                    ],
                    [y_orig, bw_height, bw_height, y_orig],
                    [y_orig, y_orig, y_orig, y_orig],
                    color="grey",
                    label="Rubble mound",
                    zorder=2,
                )

        # Crest height
        plt.axhline(
            self.dike_crest_level,
            linestyle=":",
            color="grey",
            label="Crest height",
            zorder=1,
        )

        # Legend, labels, etc
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        plt.legend(by_label.values(), by_label.keys(), loc="upper left")
        plt.title(f"{self.profile_name} ({self.dike_orientation}°)", fontweight="bold")
        plt.xlabel("Distance [m]")
        plt.ylabel("Level [NAP+m]")
        plt.show()

    @classmethod
    def from_prfl(cls, prfl_path: str, profile_name: str = "Profile"):
        """
        Import a profile from a .prfl file

        Parameters
        ----------
        prfl_path : str
            Path to the '.prfl' file
        profile_name : str
            Name of the profile (default: 'Profile')

        Returns
        -------
        Profile
            Profile object
        """
        # Check if the file extension if .prfl
        if not prfl_path.lower().endswith(".prfl"):
            raise FileNotFoundError(
                f"[ERROR] Input file: {prfl_path} should be a .prfl file."
            )

        # Check if the provided path exists
        if not os.path.exists(prfl_path):
            raise FileNotFoundError(f"[ERROR] Input file: {prfl_path} not found.")

        # Read the file
        with open(prfl_path, "r") as file:
            prfl = file.read()
            prfl = prfl.replace("\t", " ")

        # Split the SQL commands by a line break (\n) and remove empty lines, stop before 'memo'
        prfl = [entry.lower().strip() for entry in prfl.split("\n") if entry.strip()]
        prfl = prfl[: prfl.index("memo")]

        # Create a new profile
        profile = cls(profile_name)

        # Version
        version = float([entry for entry in prfl if "versie" in entry][0].split(" ")[1])
        if version != 4.0:
            raise NotImplementedError(
                f"[ERROR] Prfl version {version} is not supported."
            )

        # Sheet pile
        sheetpile = (
            True
            if float([entry for entry in prfl if "damwand" in entry][0].split(" ")[1])
            == 1.0
            else False
        )
        if sheetpile:
            raise NotImplementedError("[ERROR] Sheet piles are not implemented.")

        # Breakwater
        breakwater = Breakwater(
            int([entry for entry in prfl if "dam" in entry][0].split(" ")[1])
        )
        breakwater_level = float(
            [entry for entry in prfl if "damhoogte" in entry][0].split(" ")[1]
        )
        profile.set_breakwater(breakwater, breakwater_level)

        # Foreland
        n_foreland = int(
            [entry for entry in prfl if "voorland" in entry][0].split(" ")[1]
        )
        idx_foreland = prfl.index(f"voorland {n_foreland}") + 1
        foreland_x = []
        foreland_y = []
        for row in range(idx_foreland, idx_foreland + n_foreland):
            foreland_x.extend([float(prfl[row].split(" ")[0])])
            foreland_y.extend([float(prfl[row].split(" ")[1])])
        if n_foreland > 0:
            profile.set_foreland_geometry(foreland_x, foreland_y)

        # Dike
        n_dike = int([entry for entry in prfl if "dijk" in entry][0].split(" ")[1])
        idx_dike = prfl.index(f"dijk {n_dike}") + 1
        dike_x = []
        dike_y = []
        dike_r = []
        for row in range(idx_dike, idx_dike + n_dike):
            dike_x.extend([float(prfl[row].split(" ")[0])])
            dike_y.extend([float(prfl[row].split(" ")[1])])
            dike_r.extend([float(prfl[row].split(" ")[2])])
        profile.set_dike_geometry(dike_x, dike_y, dike_r)

        # Crest height
        dike_crest_level = float(
            [entry for entry in prfl if "kruinhoogte" in entry][0].split(" ")[1]
        )
        profile.set_dike_crest_level(dike_crest_level)

        # Dike orientation
        dike_orientation = float(
            [entry for entry in prfl if "richting" in entry][0].split(" ")[1]
        )
        profile.set_dike_orientation(dike_orientation)

        # Return the class
        return profile

    @classmethod
    def from_dictionary(cls, dictionary: dict):
        """
        Create a profile from a dictionary

        Parameters
        ----------
        dictionary: dict
            Dictionary with all profile settings
        """
        # Create a new profile
        profile = cls(dictionary["profile_name"])

        # Add settings from dictionary to profile
        for setting in dictionary.keys():
            setattr(profile, setting, dictionary[setting])

        # Validate
        profile.validate_profile()

        return profile

    @classmethod
    def from_gebugekb_tool(
        cls, sql_path: str, profile_name: str = "Profile", berm_slope: float = 1 / 100
    ):
        """
        Import a profile from the GEBUGEKB plugin.

        Parameters
        ----------
        sql_path : str
            Path to the '1.sql' file
        profile_name : str
            Name of the profile (default: 'Profile')
        berm_slope : float
            If applicable: slope of the berm (default : 1/100)

        Returns
        -------
        Profile
            Profile object
        """
        # Check if the file extension if .prfl
        if not sql_path.lower().endswith(".sql"):
            raise FileNotFoundError(
                f"[ERROR] Input file: {sql_path} should be a .sql file."
            )

        # Check if the provided path exists
        if not os.path.exists(sql_path):
            raise FileNotFoundError(f"[ERROR] Input file: {sql_path} not found.")

        # Check if the berm slope is not too steep or shallow
        if berm_slope < 1 / 100 or berm_slope > 1 / 15:
            raise ValueError(
                f"[ERROR] The slope of the berm cannot be steeper than 1/15 or shallower than 1/100 (Given: 1/{1/berm_slope})."
            )

        # Read the file
        with open(sql_path, "r") as file:
            sql_commands = file.read()

        # Split the SQL commands by a line break (\n)
        commands = sql_commands.split("\n")

        # Remove all lines starting with 'DELETE FROM', '--' or is empty. Remove comments (by splitting at the semicolumn)
        commands = [
            entry.split(";")[0]
            for entry in commands
            if not entry.startswith("DELETE FROM")
            and not entry.startswith("--")
            and not entry == ""
        ]

        # Create a new profile
        profile = cls(profile_name)

        # Breakwater
        breakwater = [entry for entry in commands if "Breakwaters" in entry]
        breakwater = np.array(
            [ast.literal_eval(entry.split("VALUES ")[-1]) for entry in breakwater]
        )
        if len(breakwater) > 1:
            raise ValueError(
                "[ERROR] Multiple breakwaters are defined. Only one can be defined at a time."
            )
        elif len(breakwater) == 1:
            profile.set_breakwater(Breakwater(int(breakwater[0][1])), breakwater[0][2])

        # Foreland
        foreland = [entry for entry in commands if "Forelands" in entry]
        foreland = np.array(
            [ast.literal_eval(entry.split("VALUES ")[-1]) for entry in foreland]
        )
        if len(foreland) > 0:
            profile.set_foreland_geometry(foreland[:, 2], foreland[:, 3])

        # Dike geometry
        dike = [
            entry.replace("NULL", "-999")
            for entry in commands
            if "VariableDatas" in entry
        ]
        dike = np.array(
            [ast.literal_eval(entry.split("VALUES ")[-1]) for entry in dike]
        )[:, 4:]

        # Dike geometry
        slope_lower = dike[dike[:, 0] == 10][0][1]
        slope_upper = dike[dike[:, 0] == 11][0][1]
        toe_level = dike[dike[:, 0] == 12][0][1]
        berm_level = dike[dike[:, 0] == 13][0][1]
        berm_length = dike[dike[:, 0] == 14][0][1]
        crest_level = dike[dike[:, 0] == 15][0][1]
        dike_orientation = dike[dike[:, 0] == 16][0][1]
        roughness = dike[dike[:, 0] == 17][0][1]
        zone_roughness = dike[dike[:, 0] == 50][0][1]
        zone_ymin = dike[dike[:, 0] == 51][0][1]
        zone_ymax = dike[dike[:, 0] == 52][0][1]

        # Calculate coordinates
        dike_x = [0]
        dike_y = [toe_level]

        # If there is no berm
        if berm_length <= 0:
            dike_x.extend([(berm_level - toe_level) / slope_lower])
            dike_y.extend([berm_level])
        else:
            berm_difference = berm_slope * berm_length
            dike_x.extend(
                [(berm_level - 0.5 * berm_difference - toe_level) / slope_lower]
            )
            dike_y.extend([berm_level - 0.5 * berm_difference])
            dike_x.extend([dike_x[-1] + berm_length])
            dike_y.extend([berm_level + 0.5 * berm_difference])

        # Crest
        dike_x.extend([dike_x[-1] + (crest_level - dike_y[-1]) / slope_upper])
        dike_y.extend([crest_level])

        # Roughness
        dike_r = np.ones([len(dike_x)]) * roughness
        dike_r[(dike_y >= zone_ymin) & (dike_y < zone_ymax)] = zone_roughness

        # Do we need to add an extra point for zone_ymin?
        if (zone_ymin >= np.min(dike_y)) and (zone_ymin not in dike_y):
            x_int = Interpolate.inextrp1d(zone_ymin, dike_y, dike_x)
            idx = np.searchsorted(dike_x, x_int)
            dike_x = np.insert(dike_x, idx, x_int)
            dike_y = np.insert(dike_y, idx, zone_ymin)
            dike_r = np.insert(dike_r, idx, zone_roughness)

        # Do we need to add an extra point for zone_ymax?
        if (zone_ymax <= np.max(dike_y)) and (zone_ymax not in dike_y):
            x_int = Interpolate.inextrp1d(zone_ymax, dike_y, dike_x)
            idx = np.searchsorted(dike_x, x_int)
            dike_x = np.insert(dike_x, idx, x_int)
            dike_y = np.insert(dike_y, idx, zone_ymax)
            dike_r = np.insert(dike_r, idx, roughness)

        # Dike orientation, crest and geometry
        profile.set_dike_orientation(dike_orientation)
        profile.set_dike_geometry(dike_x, dike_y, dike_r)
        profile.set_dike_crest_level(crest_level)

        # Return the class
        return profile



import numpy as np

from .....io.file_hydranl import FileHydraNL


class DiscreteProbability:
    """
    Class to describe the discrete statistics.
    """

    def __init__(self, statistics_file_path: str):
        """
        Constructor class for the discrete statistics.

        Parameters
        ----------
        statistics_file_path : str
            Path to the statistics file
        """
        self.discretisation, self.probability = FileHydraNL.read_file_2columns(
            statistics_file_path
        )

    def __len__(self):
        """
        Return the number discretisations.

        Returns
        -------
        int
            Number of discretisations
        """
        return len(self.discretisation)

    def get_discretisation(self) -> np.ndarray:
        """
        Return the discretisation.

        Returns
        -------
        np.ndarray
            1D array with discretisation
        """
        return self.discretisation

    def get_probability(self) -> np.ndarray:
        """
        Return the probability of each discretisation.

        Returns
        -------
        np.ndarray
            1D array with probabilities
        """
        return self.probability
